from typing import Any


def format_list_no_quotes(a: Any | list[Any]) -> str:
    if isinstance(a, list):
        return "[" + ", ".join(format_list_no_quotes(x) for x in a) + "]"  # type: ignore
    return str(a)



from functools import partial
from typing import Literal, cast

import numpy as np
import numpy.typing as npt


def bits_to_int(
    input: npt.ArrayLike,
    bit_order: Literal["LSB-first", "MSB-first"] = "LSB-first",
) -> int | npt.NDArray[np.integer]:
    r"""
    Converts a bit array to its integer representation.

    Parameters:
        input: The input bit array. Must be an array with elements in the set $\\{ 0, 1 \\}$, with the bit sequences in the last axis.

        bit_order: Bit order convention. Must be either `"LSB-first"` (least significant bit in the first position) or `"MSB-first"` (most significant bit in the first position). The default value is `"LSB-first"`.

    Returns:
        output: The integer representation of the input bit array. Has the same shape as the input, but with the last dimension removed.

    Examples:
        >>> komm.bits_to_int([0, 0, 0, 0, 1, 0], bit_order="LSB-first")
        16

        >>> komm.bits_to_int([0, 0, 0, 0, 1, 0], bit_order="MSB-first")
        2

        >>> komm.bits_to_int([[0, 0], [1, 0], [0, 1], [1, 1]])
        array([0, 1, 2, 3])

        >>> komm.bits_to_int([[0, 0], [1, 0], [0, 1], [1, 1]], bit_order="MSB-first")
        array([0, 2, 1, 3])
    """
    if np.ndim(input) == 1:
        if bit_order == "LSB-first":
            input = np.asarray(input, dtype=np.uint8)
        elif bit_order == "MSB-first":
            input = np.flip(input, axis=-1)
        else:
            raise ValueError("'bit_order' must be in {'LSB-first', 'MSB-first'}")
        packed = np.packbits(input, bitorder="little")
        return int.from_bytes(packed, byteorder="little")
    return np.apply_along_axis(partial(bits_to_int, bit_order=bit_order), -1, input)


def int_to_bits(
    input: npt.ArrayLike,
    width: int,
    bit_order: Literal["LSB-first", "MSB-first"] = "LSB-first",
) -> npt.NDArray[np.integer]:
    r"""
    Converts an integer, or array of integers, to their bit representations.

    Parameters:
        input: The input integer, or array of integers.

        width: The width of the bit representation.

        bit_order: Bit order convention. Must be either `"LSB-first"` (least significant bit in the first position) or `"MSB-first"` (most significant bit in the first position). The default value is `"LSB-first"`.

    Returns:
        output: The bit representation of the input, with the bit sequences in the last axis. Has the same shape as the input, but with a new last dimension of size `width` appended.

    Examples:
        >>> komm.int_to_bits(2, width=6, bit_order="LSB-first")
        array([0, 1, 0, 0, 0, 0])

        >>> komm.int_to_bits(2, width=6, bit_order="MSB-first")
        array([0, 0, 0, 0, 1, 0])

        >>> komm.int_to_bits([0, 1, 2, 3], width=4)
        array([[0, 0, 0, 0],
               [1, 0, 0, 0],
               [0, 1, 0, 0],
               [1, 1, 0, 0]])

        >>> komm.int_to_bits([0, 1, 2, 3], width=4, bit_order="MSB-first")
        array([[0, 0, 0, 0],
               [0, 0, 0, 1],
               [0, 0, 1, 0],
               [0, 0, 1, 1]])
    """
    if np.ndim(input) == 0:
        input = cast(int, input)
        if bit_order == "LSB-first":
            bit_list = [(input >> i) % 2 for i in range(width)]
        elif bit_order == "MSB-first":
            bit_list = [(input >> (width - 1 - i)) % 2 for i in range(width)]
        else:
            raise ValueError("'bit_order' must be in {'LSB-first', 'MSB-first'}")
        return np.array(bit_list, dtype=int)
    input = np.asarray(input, dtype=object)
    bits = np.array(
        [int_to_bits(x, width=width, bit_order=bit_order) for x in input.ravel()],
        dtype=int,
    )
    return bits.reshape(input.shape + (width,))



from abc import ABC, abstractmethod
from collections.abc import Callable
from functools import cached_property

import numpy as np
import numpy.typing as npt


class ScalarQuantizer(ABC):
    @cached_property
    @abstractmethod
    def levels(self) -> npt.NDArray[np.floating]:
        r"""
        The quantizer levels $v_0, v_1, \ldots, v_{L-1}$.
        """
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def thresholds(self) -> npt.NDArray[np.floating]:
        r"""
        The quantizer finite thresholds $t_1, t_2, \ldots, t_{L-1}$.
        """
        raise NotImplementedError

    @abstractmethod
    def digitize(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Returns the quantization indices for the input signal.

        Parameters:
            input: The input signal $x$ to be digitized.

        Returns:
            output: The integer indices of the quantization levels for each input sample.
        """
        input = np.asarray(input, dtype=float)
        output = np.digitize(input, self.thresholds, right=False)
        return output

    @abstractmethod
    def quantize(self, input: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        Quantizes the input signal.

        Parameters:
            input: The input signal $x$ to be quantized.

        Returns:
            output: The quantized signal $y$.
        """
        return self.levels[self.digitize(input)]

    @abstractmethod
    def mean_squared_error(
        self,
        input_pdf: Callable[[npt.NDArray[np.floating]], npt.NDArray[np.floating]],
        input_range: tuple[float, float],
        points_per_interval: int = 4096,
    ) -> float:
        r"""
        Computes the mean squared (quantization) error (MSE) of the quantizer for a given input pdf. It is defined as
        $$
            \mse = \int_{-\infty}^{\infty} (y - x)^2 f_X(x) \, dx
        $$
        where $y$ is the quantized signal and $f_X(x)$ is the pdf of the input signal.

        Parameters:
            input_pdf: The pdf $f_X(x)$ of the input signal.
            input_range: The range $(x_\mathrm{min}, x_\mathrm{max})$ of the input signal.
            points_per_interval: The number of points per interval for numerical integration (default: 4096).

        Returns:
            mse: The mean square quantization error.
        """
        # See [Say06, eq. (9.3)].
        x_min, x_max = input_range
        thresholds = np.concatenate(([x_min], self.thresholds, [x_max]))
        mse = 0.0
        for i, level in enumerate(self.levels):
            left, right = thresholds[i], thresholds[i + 1]
            x = np.linspace(left, right, num=points_per_interval, dtype=float)
            pdf = input_pdf(x)
            integrand: npt.NDArray[np.floating] = (level - x) ** 2 * pdf
            integral = np.trapezoid(integrand, x)
            mse += float(integral)
        return mse



from abc import ABC, abstractmethod
from typing import TypeVar

import numpy as np
import numpy.typing as npt

from ..types import Array2D

T = TypeVar("T", np.floating, np.complexfloating)


class Labeling(ABC):
    @property
    @abstractmethod
    def matrix(self) -> Array2D[np.integer]:
        r"""
        The labeling matrix $\mathbf{Q}$.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def num_bits(self) -> int:
        r"""
        The number $m$ of bits per index of the labeling.
        """
        return self.matrix.shape[1]

    @property
    @abstractmethod
    def cardinality(self) -> int:
        r"""
        The cardinality $2^m$ of the labeling.
        """
        return self.matrix.shape[0]

    @property
    @abstractmethod
    def inverse_mapping(self) -> dict[tuple[int, ...], int]:
        r"""
        The inverse mapping of the labeling. It is a dictionary that maps each binary tuple to the corresponding index.
        """
        return {tuple(map(int, row)): i for i, row in enumerate(self.matrix)}

    @abstractmethod
    def indices_to_bits(self, indices: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Returns the binary representation of the given indices.

        Parameters:
            indices: The indices to be converted to bits. Must be an array of integers in $[0:2^m)$.

        Returns:
            bits: The binary representations of the given indices. Has the same shape as `indices`, but with the last dimension expanded by a factor of $m$.
        """
        M = self.cardinality
        indices = np.asarray(indices, dtype=int)
        if not (np.all(indices >= 0) and np.all(indices < M)):
            raise ValueError(f"elements of 'index' must be in [0:{M})")
        bits = self.matrix[indices].reshape(*indices.shape[:-1], -1)
        return bits

    @abstractmethod
    def bits_to_indices(self, bits: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Returns the indices corresponding to a given sequence of bits.

        Parameters:
            bits: The bits to be converted to indices. Must be an array with elements in $\mathbb{B}$ whose last dimension is a multiple $m$.

        Returns:
            indices: The indices corresponding to the given bits. Has the same shape as `bits`, but with the last dimension contracted by a factor of $m$.
        """
        m = self.num_bits
        bits = np.asarray(bits, dtype=int)
        if bits.shape[-1] % m != 0:
            raise ValueError(
                "last dimension of 'bits' must be a multiple of the number of"
                f" bits per index {m} (got {bits.shape[-1]})"
            )
        if not np.all(np.isin(bits, [0, 1])):
            raise ValueError("elements of 'bits' must be either 0 or 1")
        indices = np.apply_along_axis(
            func1d=lambda row: self.inverse_mapping[tuple(row)],
            axis=-1,
            arr=bits.reshape(*bits.shape[:-1], -1, m),
        ).reshape(*bits.shape[:-1], -1)
        return indices

    @abstractmethod
    def marginalize(self, metrics: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        Marginalize metrics over the bits of the labeling. The metrics may represent likelihoods or probabilities, for example. The marginalization is done by computing the L-values of the bits, which are defined as
        $$
        L(\mathtt{b}_i) = \log \frac{\Pr[\mathtt{b}_i = 0]}{\Pr[\mathtt{b}_i = 1]}.
        $$

        Parameters:
            metrics: The metrics for each index of the labeling. Must be an array whose last dimension is a multiple of $2^m$.

        Returns:
            lvalues: The marginalized metrics over the bits of the labeling. Has the same shape as `metrics`, but with the last dimension changed by a factor of $m / 2^m$.
        """
        m, M = self.num_bits, self.cardinality
        metrics = np.asarray(metrics)
        if metrics.shape[-1] % M != 0:
            raise ValueError(
                "last dimension of 'metrics' must be a multiple of the cardinality"
                f" {M} of the labeling (got {metrics.shape[-1]})"
            )
        mask0 = self.matrix == 0
        mask1 = self.matrix == 1
        m = metrics.reshape(*metrics.shape[:-1], -1, M)
        with np.errstate(divide="ignore"):
            lvalues = np.log(m @ mask0) - np.log(m @ mask1)
        return lvalues.reshape(*metrics.shape[:-1], -1)



import numpy as np
import numpy.typing as npt


def zadoff_chu_sequence(
    length: int, root_index: int
) -> npt.NDArray[np.complexfloating]:
    n = np.arange(length)
    return np.exp(-1j * np.pi * root_index * n * (n + 1) / length)



from collections.abc import Callable
from functools import cached_property
from itertools import product
from typing import Any, TypedDict

import numpy as np
import numpy.typing as npt


class MetricMemory(TypedDict):
    paths: npt.NDArray[np.integer]
    metrics: npt.NDArray[np.floating]


class MealyMachine:
    r"""
    Finite-state Mealy machine. It is defined by a *set of states* $\mathcal{S}$, an *input alphabet* $\mathcal{X}$, an *output alphabet* $\mathcal{Y}$, a *transition function* $T : \mathcal{S} \times \mathcal{X} \to \mathcal{S}$, and an *output function* $G : \mathcal{S} \times \mathcal{X} \to \mathcal{Y}$. Here, for simplicity, the set of states, the input alphabet, and the output alphabet are taken as $\mathcal{S} = [0 : |\mathcal{S}|)$, $\mathcal{X} = [0 : |\mathcal{X}|)$, and $\mathcal{Y} = [0 : |\mathcal{Y}|)$, respectively. more details, see [Wikipedia: Mealy machine](https://en.wikipedia.org/wiki/Mealy_machine).


    Parameters:
        transitions: The matrix of transitions of the machine, of shape $|\mathcal{S}| \times |\mathcal{X}|$. The element in row $s \in \mathcal{S}$ and column $x \in \mathcal{X}$ should be $T(s, x) \in \mathcal{S}$, that is, the next state of the machine given that the current state is $s$ and the input is $x$.

        outputs: The matrix of outputs of the machine, of shape $|\mathcal{S}| \times |\mathcal{X}|$. The element in row $s \in \mathcal{S}$ and column $x \in \mathcal{X}$ should be $G(s, x) \in \mathcal{Y}$, that is, the output of the machine given that the current state is $s$ and the input is $x$.

    Examples:
        1. Consider the finite-state Mealy machine whose state diagram depicted in the figure below.

            <figure markdown>
            ![Finite-state Mealy machine example.](/fig/mealy.svg)
            </figure>

            It has set of states $\mathcal{S} = \\{ 0, 1, 2, 3 \\}$, input alphabet $\mathcal{X} = \\{ 0, 1 \\}$, output alphabet $\mathcal{Y} = \\{ 0, 1, 2, 3 \\}$. The transition function $T$ and output function $G$ are given by the table below.

            | State $s$ | Input $x$ | Transition $T(s, x)$ | Output $G(s, x)$ |
            | :-------: | :-------: | :------------------: | :--------------: |
            | $0$       | $0$       | $0$                  | $0$              |
            | $0$       | $1$       | $1$                  | $3$              |
            | $1$       | $0$       | $2$                  | $1$              |
            | $1$       | $1$       | $3$                  | $2$              |
            | $2$       | $0$       | $0$                  | $3$              |
            | $2$       | $1$       | $1$                  | $0$              |
            | $3$       | $0$       | $2$                  | $2$              |
            | $3$       | $1$       | $3$                  | $1$              |

                >>> machine = komm.MealyMachine(
                ...     transitions=[[0, 1], [2, 3], [0, 1], [2, 3]],
                ...     outputs=[[0, 3], [1, 2], [3, 0], [2, 1]],
                ... )
    """

    def __init__(self, transitions: npt.ArrayLike, outputs: npt.ArrayLike):
        self.transitions = np.asarray(transitions)
        self.outputs = np.asarray(outputs)

    def __repr__(self) -> str:
        args = ", ".join([
            f"transitions={self.transitions.tolist()}",
            f"outputs={self.outputs.tolist()}",
        ])
        return f"{self.__class__.__name__}({args})"

    @cached_property
    def num_states(self) -> int:
        r"""
        The number of states of the machine.

        Examples:
            >>> machine = komm.MealyMachine(
            ...     transitions=[[0, 1], [2, 3], [0, 1], [2, 3]],
            ...     outputs=[[0, 3], [1, 2], [3, 0], [2, 1]],
            ... )
            >>> machine.num_states
            4
        """
        return self.transitions.shape[0]

    @cached_property
    def num_input_symbols(self) -> int:
        r"""
        The size (cardinality) of the input alphabet $\mathcal{X}$.

        Examples:
            >>> machine = komm.MealyMachine(
            ...     transitions=[[0, 1], [2, 3], [0, 1], [2, 3]],
            ...     outputs=[[0, 3], [1, 2], [3, 0], [2, 1]],
            ... )
            >>> machine.num_input_symbols
            2
        """
        return self.transitions.shape[1]

    @cached_property
    def num_output_symbols(self) -> int:
        r"""
        The size (cardinality) of the output alphabet $\mathcal{Y}$.

        Examples:
            >>> machine = komm.MealyMachine(
            ...     transitions=[[0, 1], [2, 3], [0, 1], [2, 3]],
            ...     outputs=[[0, 3], [1, 2], [3, 0], [2, 1]],
            ... )
            >>> machine.num_output_symbols
            4
        """
        return int(np.amax(self.outputs)) + 1

    @cached_property
    def input_edges(self) -> npt.NDArray[np.integer]:
        r"""
        The matrix of input edges of the machine. It has shape $|\mathcal{S}| \times |\mathcal{S}|$. If there is an edge from $s_0 \in \mathcal{S}$ to $s_1 \in \mathcal{S}$, then the element in row $s_0$ and column $s_1$ is the input associated with that edge (an element of $\mathcal{X}$); if there is no such edge, then the element is $-1$.

        Examples:
            >>> machine = komm.MealyMachine(
            ...     transitions=[[0, 1], [2, 3], [0, 1], [2, 3]],
            ...     outputs=[[0, 3], [1, 2], [3, 0], [2, 1]],
            ... )
            >>> machine.input_edges
            array([[ 0,  1, -1, -1],
                   [-1, -1,  0,  1],
                   [ 0,  1, -1, -1],
                   [-1, -1,  0,  1]])
        """
        input_edges = np.full((self.num_states, self.num_states), fill_value=-1)
        for state_from in range(self.num_states):
            for x, state_to in enumerate(self.transitions[state_from, :]):
                input_edges[state_from, state_to] = x
        return input_edges

    @cached_property
    def output_edges(self) -> npt.NDArray[np.integer]:
        r"""
        The matrix of output edges of the machine. It has shape $|\mathcal{S}| \times |\mathcal{S}|$. If there is an edge from $s_0 \in \mathcal{S}$ to $s_1 \in \mathcal{S}$, then the element in row $s_0$ and column $s_1$ is the output associated with that edge (an element of $\mathcal{Y}$); if there is no such edge, then the element is $-1$.

        Examples:
            >>> machine = komm.MealyMachine(
            ...     transitions=[[0, 1], [2, 3], [0, 1], [2, 3]],
            ...     outputs=[[0, 3], [1, 2], [3, 0], [2, 1]],
            ... )
            >>> machine.output_edges
            array([[ 0,  3, -1, -1],
                   [-1, -1,  1,  2],
                   [ 3,  0, -1, -1],
                   [-1, -1,  2,  1]])
        """
        output_edges = np.full((self.num_states, self.num_states), fill_value=-1)
        for state_from in range(self.num_states):
            for x, state_to in enumerate(self.transitions[state_from, :]):
                output_edges[state_from, state_to] = self.outputs[state_from, x]
        return output_edges

    def process(
        self,
        input: npt.ArrayLike,
        initial_state: int,
    ) -> tuple[npt.NDArray[np.integer], int]:
        r"""
        Returns the output sequence corresponding to a given input sequence. It assumes the machine starts at a given initial state $s_\mathrm{i}$. The input sequence and the output sequence are denoted by $x = (x_0, x_1, \ldots, x_{L-1}) \in \mathcal{X}^L$ and $y = (y_0, y_1, \ldots, y_{L-1}) \in \mathcal{Y}^L$, respectively.

        Parameters:
            input: The input sequence $x \in \mathcal{X}^L$. It should be a 1D-array with elements in $\mathcal{X}$.

            initial_state: The initial state $s_\mathrm{i}$ of the machine. Should be an integer in $\mathcal{S}$.

        Returns:
            output: The output sequence $y \in \mathcal{Y}^L$ corresponding to `input`, assuming the machine starts at the state given by `initial_state`. It is a 1D-array with elements in $\mathcal{Y}$.

            final_state: The final state $s_\mathrm{f}$ of the machine. It is an integer in $\mathcal{S}$.

        Examples:
            >>> machine = komm.MealyMachine(
            ...     transitions=[[0, 1], [2, 3], [0, 1], [2, 3]],
            ...     outputs=[[0, 3], [1, 2], [3, 0], [2, 1]],
            ... )
            >>> input, initial_state = [1, 1, 0, 1, 0], 0
            >>> output, final_state = machine.process(input, initial_state)
            >>> output
            array([3, 2, 2, 0, 1])
            >>> final_state
            2
        """
        output = np.empty_like(input, dtype=int)
        s = initial_state
        for t, x in np.ndenumerate(input):
            y = self.outputs[s, x]
            s = self.transitions[s, x]
            output[t] = y
        final_state = int(s)
        return output, final_state

    def viterbi(
        self,
        observed: npt.ArrayLike,
        metric_function: Callable[[int, Any], float],
        initial_metrics: npt.ArrayLike | None = None,
    ) -> tuple[npt.NDArray[np.integer], npt.NDArray[np.floating]]:
        r"""
        Applies the Viterbi algorithm on a given observed sequence. The Viterbi algorithm finds the most probable input sequence $\hat{x} \in \mathcal{X}^L$ ending in state $s$, for all $s \in \mathcal{S}$, given an observed sequence $z \in \mathcal{Z}^L$. It is assumed uniform input priors. See <cite>LC04, Sec. 12.1</cite>.

        Parameters:
            observed: The observed sequence $z \in \mathcal{Z}^L$.

            metric_function: The metric function $\mathcal{Y} \times \mathcal{Z} \to \mathbb{R}$.

            initial_metrics: The initial metrics for each state. It must be a 1D-array of length $|\mathcal{S}|$. The default value is `0.0` for all states.

        Returns:
            input_hat: The most probable input sequence $\hat{x} \in \mathcal{X}^L$ ending in state $s$, for all $s \in \mathcal{S}$. It is a 2D-array of shape $L \times |\mathcal{S}|$, in which column $s$ is equal to $\hat{x}$.

            final_metrics: The final metrics for each state. It is a 1D-array of length $|\mathcal{S}|$.
        """
        observed = np.asarray(observed)
        L, num_states = observed.shape[0], self.num_states
        choices = np.zeros((L, num_states), dtype=int)
        metrics = np.full((L + 1, num_states), fill_value=np.inf)
        if initial_metrics is None:
            metrics[0, :] = np.zeros(num_states, dtype=float)
        else:
            metrics[0, :] = initial_metrics
        for t, z in enumerate(observed):
            for s0 in range(num_states):
                for s1, y in zip(self.transitions[s0], self.outputs[s0]):
                    candidate_metrics = metrics[t, s0] + metric_function(y, z)
                    if candidate_metrics < metrics[t + 1, s1]:
                        metrics[t + 1, s1] = candidate_metrics
                        choices[t, s1] = s0

        # Backtrack
        input_hat = np.empty((L, num_states), dtype=int)
        for final_state in range(num_states):
            s1 = final_state
            for t in reversed(range(L)):
                s0 = choices[t, s1]
                input_hat[t, final_state] = self.input_edges[s0, s1]
                s1 = s0

        return input_hat, metrics[L, :]

    def viterbi_streaming(
        self,
        observed: npt.ArrayLike,
        metric_function: Callable[[int, Any], float],
        memory: MetricMemory,
    ) -> npt.NDArray[np.integer]:
        r"""
        Applies the streaming version of the Viterbi algorithm on a given observed sequence. The path memory (or traceback length) is denoted by $\tau$. It chooses the survivor with best metric and selects the information block on this path. See <cite>LC04, Sec. 12.3</cite>.

        Parameters:
            observed: The observed sequence $z \in \mathcal{Z}^L$.

            metric_function: The metric function $\mathcal{Y} \times \mathcal{Z} \to \mathbb{R}$.

            memory: The metrics for each state. It must be a dictionary containing two keys: `'paths'`, a 2D-array of integers of shape $|\mathcal{S}| \times (\tau + 1)$; and `'metrics'`, a 2D-array of floats of shape $|\mathcal{S}| \times (\tau + 1)$. This dictionary is updated in-place by this method.

        Returns:
            input_hat: The most probable input sequence $\hat{x} \in \mathcal{X}^L$
        """
        observed = np.asarray(observed)
        L, num_states = observed.shape[0], self.num_states
        input_hat = np.empty(L, dtype=int)
        for t, z in enumerate(observed):
            new_metrics = np.full(num_states, fill_value=np.inf)
            choices = np.zeros(num_states, dtype=int)
            for s0 in range(num_states):
                for s1, y in zip(self.transitions[s0], self.outputs[s0]):
                    candidate_metric = memory["metrics"][s0, -1] + metric_function(y, z)
                    if candidate_metric < new_metrics[s1]:
                        new_metrics[s1] = candidate_metric
                        choices[s1] = s0

            s_star = np.argmin(new_metrics)
            s0, s1 = memory["paths"][s_star, :2]
            input_hat[t] = self.input_edges[s0, s1]

            memory["metrics"] = np.roll(memory["metrics"], shift=-1, axis=1)
            memory["metrics"][:, -1] = new_metrics
            memory["paths"] = np.roll(memory["paths"], shift=-1, axis=1)

            paths_copy = np.copy(memory["paths"])
            for s1, s0 in enumerate(choices):
                memory["paths"][s1, :-1] = paths_copy[s0, :-1]
                memory["paths"][s1, -1] = s1

        return input_hat

    def forward_backward(
        self,
        observed: npt.ArrayLike,
        metric_function: Callable[[int, Any], float],
        input_priors: npt.ArrayLike | None = None,
        initial_state_distribution: npt.ArrayLike | None = None,
        final_state_distribution: npt.ArrayLike | None = None,
    ) -> npt.NDArray[np.floating]:
        r"""
        Applies the forward-backward algorithm on a given observed sequence. The forward-backward algorithm computes the posterior pmf of each input $x_0, x_1, \ldots, x_{L-1} \in \mathcal{X}$ given an observed sequence $z = (z_0, z_1, \ldots, z_{L-1}) \in \mathcal{Z}^L$. The prior pmf of each input may also be provided. See <cite>LC04, 12.6</cite>.

        Parameters:
            observed: The observed sequence $z \in \mathcal{Z}^L$.

            metric_function: The metric function $\mathcal{Y} \times \mathcal{Z} \to \mathbb{R}$.

            input_priors: The prior pmf of each input, of shape $L \times |\mathcal{X}|$. The element in row $t \in [0 : L)$ and column $x \in \mathcal{X}$ should be $p(x_t = x)$. The default value yields uniform priors.

            initial_state_distribution: The pmf of the initial state of the machine. It must be a 1D-array of length $|\mathcal{S}|$. The default value is uniform over all states.

            final_state_distribution: The pmf of the final state of the machine. It must be a 1D-array of length $|\mathcal{S}|$. The default value is uniform over all states.

        Returns:
            input_posteriors: The posterior pmf of each input, given the observed sequence, of shape $L \times |\mathcal{X}|$. The element in row $t \in [0 : L)$ and column $x \in \mathcal{X}$ is $p(x_t = x \mid z)$.
        """
        observed = np.asarray(observed)
        L, num_states = observed.shape[0], self.num_states
        num_input_symbols = self.num_input_symbols

        if input_priors is None:
            input_priors = np.ones((L, num_input_symbols)) / num_input_symbols
        if initial_state_distribution is None:
            initial_state_distribution = np.ones(num_states) / num_states
        if final_state_distribution is None:
            final_state_distribution = np.ones(num_states) / num_states

        log_gamma = np.full((L, num_states, num_states), fill_value=-np.inf)
        log_alpha = np.full((L + 1, num_states), fill_value=-np.inf)
        log_beta = np.full((L + 1, num_states), fill_value=-np.inf)

        with np.errstate(divide="ignore"):
            log_input_priors = np.log(input_priors)
            log_alpha[0, :] = np.log(initial_state_distribution)
            log_beta[L, :] = np.log(final_state_distribution)

        for t, z in enumerate(observed):
            for x, s0 in product(range(num_input_symbols), range(num_states)):
                y, s1 = self.outputs[s0, x], self.transitions[s0, x]
                log_gamma[t, s0, s1] = log_input_priors[t, x] + metric_function(y, z)

        for t in range(0, L - 1):
            for s1 in range(num_states):
                log_alpha[t + 1, s1] = np.logaddexp.reduce(
                    log_gamma[t, :, s1] + log_alpha[t, :]
                )

        for t in range(L - 1, -1, -1):
            for s0 in range(num_states):
                log_beta[t, s0] = np.logaddexp.reduce(
                    log_gamma[t, s0, :] + log_beta[t + 1, :]
                )

        log_input_posteriors = np.empty((L, num_input_symbols), dtype=float)
        edge_labels = np.empty(num_states, dtype=float)
        for t in range(L):
            for x in range(num_input_symbols):
                for s0 in range(num_states):
                    s1 = self.transitions[s0, x]
                    edge_labels[s0] = (
                        log_alpha[t, s0] + log_gamma[t, s0, s1] + log_beta[t + 1, s1]
                    )
                log_input_posteriors[t, x] = np.logaddexp.reduce(edge_labels)

        input_posteriors = np.exp(
            log_input_posteriors - np.amax(log_input_posteriors, axis=1, keepdims=True)
        )
        input_posteriors /= np.sum(input_posteriors, axis=1, keepdims=True)

        return input_posteriors



from dataclasses import dataclass
from functools import cached_property

import numpy as np
import numpy.typing as npt

from .. import abc
from .util import raised_cosine, rect


@dataclass
class RaisedCosinePulse(abc.Pulse):
    r"""
    Raised-cosine pulse. For a given *roll-off factor* $\alpha$ satisfing $0 \leq \alpha \leq 1$, it is a [pulse](/ref/Pulse)
    with spectrum given by
    $$
        \hat{p}(f) = \begin{cases}
            1, & |f| \leq f_1, \\\\[1ex]
            \dfrac{1}{2} \left( 1 + \cos \left( \pi \dfrac{|f| - f_1}{f_2 - f_1}\right) \right), & f_1 \leq |f| \leq f_2, \\\\[1ex]
            0, & \text{otherwise}.
        \end{cases}
    $$
    where $f_1 = (1 - \alpha) / 2$ and $f_2 = (1 + \alpha) / 2$.

    The waveform of the raised-cosine pulse is depicted below for $\alpha = 0.25$, and for $\alpha = 0.75$.

    <div class="centered" markdown>
      <span>
      ![Raised-cosine pulse with roll-off factor 0.25.](/fig/pulse_raised_cosine_25.svg)
      </span>
      <span>
      ![Raised-cosine pulse with roll-off factor 0.75.](/fig/pulse_raised_cosine_75.svg)
      </span>
    </div>

    For more details, see [Wikipedia: Raised-cosine filter](https://en.wikipedia.org/wiki/Raised-cosine_filter).

    Notes:
        - For $\alpha = 0$ it reduces to the [sinc pulse](/ref/SincPulse).
        - For $\alpha = 1$ it becomes what is known as the _full cosine roll-off pulse_.

    Parameters:
        rolloff: The roll-off factor $\alpha$ of the pulse. Must satisfy $0 \leq \alpha \leq 1$. The default value is `1.0`.
    """

    rolloff: float = 1.0

    def waveform(self, t: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        For the raised-cosine pulse, it is given by
        $$
            p(t) = \sinc(t) \frac{\cos(\pi \alpha t)}{1 - (2 \alpha t)^2}.
        $$

        Examples:
            >>> pulse = komm.RaisedCosinePulse(rolloff=0.25)
            >>> pulse.waveform(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... ).round(3)
            array([0.   , 0.29 , 0.627, 0.897, 1.   , 0.897, 0.627, 0.29 , 0.   ])
        """
        return raised_cosine(t, self.rolloff)

    def spectrum(self, f: npt.ArrayLike) -> npt.NDArray[np.complexfloating]:
        r"""
        Examples:
            >>> pulse = komm.RaisedCosinePulse(rolloff=0.25)
            >>> np.abs(pulse.spectrum(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... ))
            array([0. , 0. , 0.5, 1. , 1. , 1. , 0.5, 0. , 0. ])
        """
        α = self.rolloff
        f = np.asarray(f)
        if α == 0:
            return rect(f).astype(complex)
        f1 = (1 - α) / 2
        f2 = (1 + α) / 2
        band1 = abs(f) < f1
        band2 = (f1 <= abs(f)) * (abs(f) < f2)
        spectrum = np.zeros_like(f, dtype=complex)
        spectrum[band1] = 1.0
        spectrum[band2] = 0.5 * (1 + np.cos(np.pi * (abs(f[band2]) - f1) / α))
        return spectrum

    def autocorrelation(self, tau: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        For the raised-cosine pulse, it is given by
        $$
            R(\tau) = \sinc(\tau) \frac{\cos(\pi \alpha \tau)}{1 - (2 \alpha \tau)^2} - \frac{\alpha}{4} \sinc(\alpha \tau) \frac{\cos(\pi \tau)}{1 - (\alpha \tau)^2}.
        $$

        Examples:
            >>> pulse = komm.RaisedCosinePulse(rolloff=0.25)
            >>> pulse.autocorrelation(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... ).round(3)
            array([0.06 , 0.334, 0.627, 0.853, 0.938, 0.853, 0.627, 0.334, 0.06 ])
        """
        α = self.rolloff
        tau = np.asarray(tau)
        if α == 0:
            return np.sinc(tau)
        with np.errstate(divide="ignore", invalid="ignore"):
            term = np.sinc(α * tau) * np.cos(np.pi * tau) / (1 - (α * tau) ** 2)
        singularity = np.isclose(np.abs(tau), 1 / α)
        term[singularity] = np.cos(np.pi / α) / 2
        return raised_cosine(tau, α) - α / 4 * term

    def energy_density_spectrum(self, f: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        For the raised-cosine pulse, it is given by
        $$
            S(f) = \begin{cases}
                1, & |f| \leq f_1, \\\\[1ex]
                \dfrac{1}{4} \left( 1 + \cos \left( \pi \dfrac{|f| - f_1}{f_2 - f_1}\right) \right)^2, & f_1 \leq |f| \leq f_2, \\\\[1ex]
                0, & \text{otherwise}.
            \end{cases}
        $$

        Examples:
            >>> pulse = komm.RaisedCosinePulse(rolloff=0.25)
            >>> pulse.energy_density_spectrum(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... )
            array([0.  , 0.  , 0.25, 1.  , 1.  , 1.  , 0.25, 0.  , 0.  ])
        """
        return np.abs(self.spectrum(f)) ** 2

    @cached_property
    def support(self) -> tuple[float, float]:
        r"""
        For the raised-cosine pulse, the support is given by $(-\infty, \infty)$.

        Examples:
            >>> pulse = komm.RaisedCosinePulse(rolloff=0.25)
            >>> pulse.support
            (-inf, inf)
        """
        return (-np.inf, np.inf)

    def taps(
        self,
        samples_per_symbol: int,
        span: tuple[int, int] | None = None,
    ) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> pulse = komm.RaisedCosinePulse(rolloff=0.25)
            >>> pulse.taps(samples_per_symbol=4, span=(-1, 1)).round(3)
            array([0.   , 0.29 , 0.627, 0.897, 1.   , 0.897, 0.627, 0.29 , 0.   ])
            >>> pulse.taps(samples_per_symbol=4, span=(-16, 16)).shape
            (129,)
        """
        return super().taps(samples_per_symbol, span)



from .ComplexSequence import ComplexSequence
from .sequences import zadoff_chu_sequence


class ZadoffChuSequence(ComplexSequence):
    r"""
    Zadoff–Chu sequence. It is a periodic, [complex sequence](/ref/ComplexSequence) given by
    $$
        z_{L,q}[n] = \mathrm{e}^{-\mathrm{j} \pi q n (n + 1) / L},
    $$
    where $L$ is the *length* (and period) of the sequence (which must be an odd integer) and $q \in [1:L)$ is called the *root index* of the sequence.

    Zadoff–Chu sequences have the following properties:

    1. *Constant amplitude:* The magnitude of the sequence satisfies
    $$
        |z_{L,q}[n]| = 1, \quad \forall n.
    $$

    2. *Zero autocorrelation:* If $q$ is coprime to $L$, then the cyclic autocorrelation of $z_{L,q}$ satisfies
    $$
        \tilde{R}\_{z_{L,q}}[\ell] = 0, \quad \forall \ell \neq 0 \mod L.
    $$

    3. *Constant cross-correlation:* If $|q' - q|$ is coprime to $L$, then the magnitude of the cyclic cross-correlation of $z_{L,q}$ and $z_{L,q'}$ satisfies
    $$
        |\tilde{R}\_{z_{L,q}, z_{L,q'}}[\ell]| = \sqrt{L}, \quad \forall \ell.
    $$

    For more details, see <cite>And22</cite>.

    Notes:
        - Theses sequences are also called *Frank–Zadoff–Chu* sequences.

    Parameters:
        length: The length $L$ of the Zadoff–Chu sequence. Must be an odd integer.
        root_index: The root index $q$ of the Zadoff–Chu sequence. Must be in $[1:L)$. The default value is $1$.

    Examples:
        >>> zadoff_chu = komm.ZadoffChuSequence(5, root_index=1)
        >>> zadoff_chu.sequence.round(6)
        array([ 1.      +0.j      ,  0.309017-0.951057j, -0.809017+0.587785j,  0.309017-0.951057j,  1.      +0.j      ])
        >>> zadoff_chu.cyclic_autocorrelation(normalized=True).round(15) + 0.0
        array([1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j, 0.+0.j])
    """

    def __init__(self, length: int, root_index: int = 1) -> None:
        if length % 2 == 0:
            raise ValueError("'length' must be an odd integer")
        if not 1 <= root_index < length:
            raise ValueError("'root_index' must be in [1 : length)")
        super().__init__(sequence=zadoff_chu_sequence(length, root_index))



from functools import cache

import numpy as np
import numpy.typing as npt

from .. import abc
from ..types import Array1D, Array2D


class PSKConstellation(abc.Constellation[np.complexfloating]):
    r"""
    Phase-shift keying (PSK) constellation. It is a complex one-dimensional [constellation](/ref/Constellation) in which the symbols are *uniformly arranged* in a circle. More precisely, the $i$-th symbol is given by
    $$
        x_i = A \exp \left( \mathrm{j} \frac{2 \pi i}{M} \right) \exp(\mathrm{j} 2 \pi \phi), \quad i \in [0 : M),
    $$
    where $M$ is the *order*, $A$ is the *amplitude*, and $\phi$ is the *phase offset* of the constellation.

    Parameters:
        order: The order $M$ of the constellation.

        amplitude: The amplitude $A$ of the constellation. The default value is `1.0`.

        phase_offset: The phase offset $\phi$ of the constellation (in turns, not radians). The default value is `0.0`.

    Examples:
        1. The $4$-PSK constellation with amplitude $A = 1$ and phase offset $\phi = 0$ is depicted below.
            <figure markdown>
            ![4-PSK constellation.](/fig/constellation_psk_4.svg)
            </figure>

                >>> const = komm.PSKConstellation(4)

        1. The $8$-PSK constellation with amplitude $A = 0.5$ and phase offset $\phi = 1/16$ is depicted below.
            <figure markdown>
            ![8-PSK constellation.](/fig/constellation_psk_8.svg)
            </figure>

                >>> const = komm.PSKConstellation(8, amplitude=0.5, phase_offset=1 / 16)
    """

    def __init__(
        self, order: int, amplitude: float = 1.0, phase_offset: float = 0.0
    ) -> None:
        self._order = order
        self._amplitude = np.float64(amplitude)
        self._phase_offset = np.float64(phase_offset)

    def __repr__(self) -> str:
        args = ", ".join([
            f"order={self._order}",
            f"amplitude={self._amplitude}",
            f"phase_offset={self._phase_offset}",
        ])
        return f"{self.__class__.__name__}({args})"

    @property
    @cache
    def matrix(self) -> Array2D[np.complexfloating]:
        r"""
        Examples:
            >>> const = komm.PSKConstellation(4)
            >>> const.matrix
            array([[ 1.+0.j],
                   [ 0.+1.j],
                   [-1.+0.j],
                   [ 0.-1.j]])
        """
        M, A, φ = self._order, self._amplitude, self._phase_offset
        i = np.arange(M).reshape(-1, 1)
        matrix = A * np.exp(2j * np.pi * i / M) * np.exp(2j * np.pi * φ)
        # We round to avoid sin(pi) != 0, and add 0.0 to avoid -0.0.
        return matrix.round(15) + 0.0

    @property
    def order(self) -> int:
        r"""
        Examples:
            >>> const = komm.PSKConstellation(4)
            >>> const.order
            4
        """
        return self._order

    @property
    def dimension(self) -> int:
        r"""
        For the PSK constellation, it is given by $N = 1$.

        Examples:
            >>> const = komm.PSKConstellation(4)
            >>> const.dimension
            1
        """
        return 1

    def mean(self, priors: npt.ArrayLike | None = None) -> Array1D[np.complexfloating]:
        r"""
        For uniform priors, the mean of the PSK constellation is given by
        $$
            \mathbf{m} = 0.
        $$

        Examples:
            >>> const = komm.PSKConstellation(4)
            >>> const.mean()
            array([0.+0.j])
        """
        if priors is None:
            return np.zeros((1,), dtype=complex)
        return super().mean(priors)

    def mean_energy(self, priors: npt.ArrayLike | None = None) -> np.floating:
        r"""
        For uniform priors, the mean energy of the PSK constellation is given by
        $$
            E = A^2.
        $$

        Examples:
            >>> const = komm.PSKConstellation(4)
            >>> const.mean_energy()
            np.float64(1.0)
        """
        if priors is None:
            A = self._amplitude
            return A**2
        return super().mean_energy(priors)

    def minimum_distance(self) -> np.floating:
        r"""
        For the PSK constellation, the minimum distance is given by
        $$
            d_\mathrm{min} = 2A\sin\left(\frac{\pi}{M}\right).
        $$

        Examples:
            >>> const = komm.PSKConstellation(4)
            >>> const.minimum_distance()  # doctest: +FLOAT_CMP
            np.float64(1.414213562373095)
        """
        M, A = self._order, self._amplitude
        return 2 * A * np.sin(np.pi / M)

    def indices_to_symbols(
        self, indices: npt.ArrayLike
    ) -> npt.NDArray[np.complexfloating]:
        r"""
        Examples:
            >>> const = komm.PSKConstellation(4)
            >>> const.indices_to_symbols([3, 0])
            array([0.-1.j, 1.+0.j])
        """
        return super().indices_to_symbols(indices)

    def closest_indices(self, received: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> const = komm.PSKConstellation(4)
            >>> const.closest_indices([0.1 - 1.1j, 1.2 + 0.1j])
            array([3, 0])
        """
        M, φ = self._order, self._phase_offset
        turn = np.angle(np.multiply(received, np.exp(-2j * np.pi * φ))) / (2 * np.pi)
        indices = np.mod(np.around(M * turn), M).astype(int)
        return indices

    def closest_symbols(
        self, received: npt.ArrayLike
    ) -> npt.NDArray[np.complexfloating]:
        r"""
        Examples:
            >>> const = komm.PSKConstellation(4)
            >>> const.closest_symbols([0.1 - 1.1j, 1.2 + 0.1j])
            array([0.-1.j, 1.+0.j])
        """
        return super().closest_symbols(received)

    def posteriors(
        self,
        received: npt.ArrayLike,
        snr: float,
        priors: npt.ArrayLike | None = None,
    ) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> const = komm.PSKConstellation(4)
            >>> const.posteriors([0.1 - 1.1j, 1.2 + 0.1j], snr=2.0).round(3)
            array([0.018, 0.   , 0.008, 0.974, 0.982, 0.012, 0.   , 0.005])
        """
        return super().posteriors(received, snr, priors)



from dataclasses import dataclass
from typing import Any

import numpy as np
import numpy.typing as npt

from .. import abc
from .._algebra.BinaryPolynomial import BinaryPolynomial
from .._algebra.FiniteBifield import FiniteBifield, FiniteBifieldElement, find_roots
from .._error_control_block.BCHCode import BCHCode
from .._util.decorators import blockwise, vectorize, with_pbar
from .util import get_pbar


@dataclass
class BerlekampDecoder(abc.BlockDecoder[BCHCode]):
    r"""
    Berlekamp decoder for [BCH codes](/ref/BCHCode). For more details, see <cite>LC04, Sec. 6.3</cite>.

    Parameters:
        code: The BCH code to be used for decoding.

    Notes:
        - Input type: `hard` (bits).
        - Output type: `hard` (bits).
    """

    code: BCHCode

    def decode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer | np.floating]:
        r"""
        Examples:
            >>> code = komm.BCHCode(4, 7)
            >>> decoder = komm.BerlekampDecoder(code)
            >>> decoder.decode([0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0])
            array([0, 0, 0, 0, 0])
        """

        @blockwise(self.code.length)
        @vectorize
        @with_pbar(get_pbar(np.size(input) // self.code.length, "Berlekamp"))
        def decode(r: npt.NDArray[np.integer]):
            r_poly = BinaryPolynomial.from_coefficients(r)
            syndrome = self.code.bch_syndrome(r_poly)
            if all(x == self.code.field.zero for x in syndrome):
                return self.code.project_word(r)
            sigma_poly = berlekamp_algorithm(self.code, syndrome)
            roots = find_roots(self.code.field, sigma_poly)
            e_loc = [e.inverse().logarithm(self.code.alpha) for e in roots]
            e_hat = np.bincount(e_loc, minlength=self.code.length)
            v_hat = (r + e_hat) % 2
            u_hat = self.code.project_word(v_hat)
            return u_hat

        return decode(input)


def berlekamp_algorithm(
    code: BCHCode, syndrome: list[Any]
) -> list[FiniteBifieldElement[FiniteBifield]]:
    # Berlekamp's iterative procedure for finding the error-location polynomial of a BCH code.
    # See [LC04, Sec. 6.3].
    field = code.field
    delta = code.delta
    sigma = {-1: [field.one], 0: [field.one]}
    discrepancy = {-1: field.one, 0: syndrome[0]}
    degree = {-1: 0, 0: 0}

    # In [LC04]: μ <-> j and ρ <-> k.
    for j in range(delta - 1):
        if discrepancy[j] == field.zero:
            degree[j + 1] = degree[j]
            sigma[j + 1] = sigma[j]
        else:
            k, max_so_far = -1, -1
            for i in range(-1, j):
                if discrepancy[i] != field.zero and i - degree[i] > max_so_far:
                    k, max_so_far = i, i - degree[i]
            degree[j + 1] = max(degree[j], degree[k] + j - k)
            fst = [field.zero] * (degree[j + 1] + 1)
            fst[: degree[j] + 1] = sigma[j]
            snd = [field.zero] * (degree[j + 1] + 1)
            snd[j - k : degree[k] + j - k + 1] = sigma[k]
            # See [LC04, eq. (6.25)].
            sigma[j + 1] = [
                fst[i] + snd[i] * discrepancy[j] / discrepancy[k]
                for i in range(degree[j + 1] + 1)
            ]
        if j < delta - 2:
            discrepancy[j + 1] = syndrome[j + 1]
            for i in range(degree[j + 1]):
                discrepancy[j + 1] += sigma[j + 1][i + 1] * syndrome[j - i]

    return sigma[delta - 1]



from functools import cache, reduce
from math import gcd

import numpy as np
import numpy.typing as npt

from .._util import global_rng
from .._util.validators import validate_transition_matrix
from ..types import Array1D, Array2D


class MarkovChain:
    r"""
    Finite-state homogeneous discrete-time Markov chain. It is defined by a finite *set of states* $\mathcal{S}$ and a *transition matrix* $P$ over $\mathcal{S}$. Here, for simplicity, the set of states is taken as $\mathcal{S} = [0 : |\mathcal{S}|)$, where $|\mathcal{S}|$ denotes the cardinality of the set of states. For more details, see <cite>YG14, Ch. MCS</cite> and <cite>GS97, Ch. 11</cite>.

    Parameters:
        transition_matrix: The transition matrix $P$ over the set of states $\mathcal{S}$. It must be a $|\mathcal{S}| \times |\mathcal{S}|$-matrix where $P_{i,j}$ is the probability of transitioning from state $i \in \mathcal{S}$ to state $j \in \mathcal{S}$. Its elements must be non-negative and each row must sum to $1$.

    Examples:
        1. Consider the finite-state homogeneous discrete-time Markov chain depicted in the figure below.

            <figure markdown>
            ![Finite-state homogeneous discrete-time Markov chain example.](/fig/markov.svg)
            </figure>

            It has set of states $\mathcal{S} = \\{ 0, 1, 2 \\}$ and transition matrix
            $$
            P = \begin{bmatrix}
                1/2 & 1/4 & 1/4 \\\\
                1/2 &  0  & 1/2 \\\\
                1/4 & 1/4 & 1/2
            \end{bmatrix}.
            $$

                >>> chain = komm.MarkovChain([
                ...     [1/2, 1/4, 1/4],
                ...     [1/2,   0, 1/2],
                ...     [1/4, 1/4, 1/2],
                ... ])
    """

    def __init__(
        self,
        transition_matrix: npt.ArrayLike,
        rng: np.random.Generator | None = None,
    ):
        self._transition_matrix = validate_transition_matrix(
            transition_matrix, square=True
        )
        self._rng = rng or global_rng.get()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.transition_matrix.tolist()})"

    @property
    def transition_matrix(self) -> npt.NDArray[np.floating]:
        return self._transition_matrix

    @property
    def num_states(self) -> int:
        r"""
        The number of states in the Markov chain, $|\mathcal{S}|$.

        Examples:
            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/4, 1/4],
            ...     [1/2,   0, 1/2],
            ...     [1/4, 1/4, 1/2],
            ... ])
            >>> chain.num_states
            3
        """
        return self._transition_matrix.shape[0]

    @cache
    def accessible_states_from(self, state: int) -> set[int]:
        r"""
        Computes the subset of states that are accessible from a given state. State $j \in \mathcal{S}$ is *accessible from* state $i \in \mathcal{S}$, denoted by $i \to j$, if there exists $n \geq 0$ such that $(P^n)_{i,j} > 0$.

        Parameters:
            state: A state $i \in \mathcal{S}$.

        Returns:
            The subset of all states accessible from state $i$.

        Examples:
            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/4, 1/4],
            ...     [1/2,   0, 1/2],
            ...     [1/4, 1/4, 1/2],
            ... ])
            >>> chain.accessible_states_from(1)
            {0, 1, 2}

            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/2,   0],
            ...     [1/2, 1/2,   0],
            ...     [  0,   0,   1],
            ... ])
            >>> chain.accessible_states_from(0)
            {0, 1}
            >>> chain.accessible_states_from(2)
            {2}
        """
        adjacency_matrix = self.transition_matrix > 0
        accessible: set[int] = set()
        stack: list[int] = [state]
        while stack:
            state = stack.pop()
            if state in accessible:
                continue
            accessible.add(state)
            neighbors = np.flatnonzero(adjacency_matrix[state])
            stack.extend(int(j) for j in neighbors if j not in accessible)
        return accessible

    @cache
    def communicating_classes(self) -> list[set[int]]:
        r"""
        Computes the communicating classes of the Markov chain. A *communicating class* is a subset of states such that every state in the class is accessible from every other state in the class. In other words, two states $i, j \in \mathcal{S}$ are in the same communicating class if and only if $i \to j$ and $j \to i$. The set of all communicating classes forms a partition of $\mathcal{S}$.

        Returns:
            A list of all communicating classes in the Markov chain.

        Examples:
            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/4, 1/4],
            ...     [1/2,   0, 1/2],
            ...     [1/4, 1/4, 1/2],
            ... ])
            >>> chain.communicating_classes()
            [{0, 1, 2}]

            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/2,   0],
            ...     [1/2, 1/2,   0],
            ...     [  0,   0,   1],
            ... ])
            >>> chain.communicating_classes()
            [{0, 1}, {2}]
        """
        reach = [self.accessible_states_from(i) for i in range(self.num_states)]
        classes: list[set[int]] = []
        visited: set[int] = set()
        for i in range(self.num_states):
            if i in visited:
                continue
            eq = {j for j in range(self.num_states) if j in reach[i] and i in reach[j]}
            classes.append(eq)
            visited.update(eq)
        return classes

    @cache
    def is_irreducible(self) -> bool:
        r"""
        Returns whether the Markov chain is irreducible. A Markov chain is *irreducible* if $i \to j$ for all $i, j \in \mathcal{S}$. Equivalently, a Markov chain is irreducible if it has only one communicating class.

        Returns:
            `True` if the Markov chain is irreducible, `False` otherwise.

        Examples:
            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/4, 1/4],
            ...     [1/2,   0, 1/2],
            ...     [1/4, 1/4, 1/2],
            ... ])
            >>> chain.is_irreducible()
            True

            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/2,   0],
            ...     [1/2, 1/2,   0],
            ...     [  0,   0,   1],
            ... ])
            >>> chain.is_irreducible()
            False

            >>> chain = komm.MarkovChain([
            ...     [0, 1, 0],
            ...     [0, 0, 1],
            ...     [1, 0, 0],
            ... ])
            >>> chain.is_irreducible()
            True
        """
        return len(self.communicating_classes()) == 1

    @cache
    def stationary_distribution(self) -> Array1D[np.floating]:
        r"""
        Computes the stationary distribution of an irreducible Markov chain. The *stationary distribution* $\pi$ is a pmf over $\mathcal{S}$ such that $\pi P = \pi$.

        Note:
            This method is only implemented for irreducible chains.

        Returns:
            The stationary distribution $\pi$ of the Markov chain.

        Examples:
            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/4, 1/4],
            ...     [1/2,   0, 1/2],
            ...     [1/4, 1/4, 1/2],
            ... ])
            >>> chain.stationary_distribution()
            array([0.4, 0.2, 0.4])

            >>> chain = komm.MarkovChain([
            ...     [0, 1, 0],
            ...     [0, 0, 1],
            ...     [1, 0, 0],
            ... ])
            >>> chain.stationary_distribution()
            array([0.33333333, 0.33333333, 0.33333333])

            >>> chain = komm.MarkovChain([
            ...     [1, 0, 0],
            ...     [0, 1, 0],
            ...     [0, 0, 1],
            ... ])
            >>> chain.stationary_distribution()
            Traceback (most recent call last):
            ...
            NotImplementedError: method is only implemented for irreducible chains
        """
        if not self.is_irreducible():
            raise NotImplementedError(
                "method is only implemented for irreducible chains"
            )
        P, n = self.transition_matrix, self.num_states
        A = np.hstack((P - np.eye(n), np.ones((n, 1))))
        b = np.concatenate((np.zeros(n), [1]))
        pi, *_ = np.linalg.lstsq(A.T, b, rcond=None)
        return pi

    @cache
    def transient_states(self) -> set[int]:
        r"""
        Returns the subset $\mathcal{T} \subseteq \mathcal{S}$ of transient states of the Markov chain. A state $i \in \mathcal{S}$ is *transient* if there exists a state $j \in \mathcal{S}$ such that $i \to j$ but $j \not\to i$.

        Returns:
            The subset $\mathcal{T}$ of transient states in the Markov chain.

        Examples:
            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/4, 1/4],
            ...     [1/2,   0, 1/2],
            ...     [1/4, 1/4, 1/2],
            ... ])
            >>> chain.transient_states()
            set()

            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/2,   0],
            ...     [1/2,   0, 1/2],
            ...     [  0,   0,   1],
            ... ])
            >>> chain.transient_states()
            {0, 1}
        """
        reach = [self.accessible_states_from(i) for i in range(self.num_states)]
        transient: set[int] = set()
        for i, r in enumerate(reach):
            for j in r:
                if i not in reach[j]:
                    transient.add(i)
                    break
        return transient

    @cache
    def recurrent_states(self) -> set[int]:
        r"""
        Returns the subset $\mathcal{R} \subseteq \mathcal{S}$ of recurrent states of the Markov chain. A state $i \in \mathcal{S}$ is *recurrent* if it is not transient.

        Returns:
            The subset $\mathcal{R}$ of recurrent states in the Markov chain.

        Examples:
            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/4, 1/4],
            ...     [1/2,   0, 1/2],
            ...     [1/4, 1/4, 1/2],
            ... ])
            >>> chain.recurrent_states()
            {0, 1, 2}

            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/2,   0],
            ...     [1/2,   0, 1/2],
            ...     [  0,   0,   1],
            ... ])
            >>> chain.recurrent_states()
            {2}
        """
        return set(range(self.num_states)) - self.transient_states()

    @cache
    def is_regular(self) -> bool:
        r"""
        Returns whether the Markov chain is regular. A Markov chain is *regular* if there exists a positive integer $n$ such that all entries of $P^n$ are positive. A finite-state Markov chain is regular if and only if it is irreducible and aperiodic.

        Returns:
            `True` if the Markov chain is regular, `False` otherwise.

        Examples:
            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/4, 1/4],
            ...     [1/2,   0, 1/2],
            ...     [1/4, 1/4, 1/2],
            ... ])
            >>> chain.is_regular()
            True

            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/2,   0],
            ...     [1/2, 1/2,   0],
            ...     [  0,   0,   1],
            ... ])
            >>> chain.is_regular()
            False

            >>> chain = komm.MarkovChain([
            ...     [0, 1, 0],
            ...     [0, 0, 1],
            ...     [1, 0, 0],
            ... ])
            >>> chain.is_regular()
            False
        """
        return self.is_irreducible() and self.is_aperiodic()

    def index_of_primitivity(self) -> int:
        r"""
        Computes the index of primitivity of a regular Markov chain. The *index of primitivity* is the smallest positive integer $n$ such that all entries of $P^n$ are positive.

        Notes:
            - This method only applies to regular Markov chains.

        Returns:
            The index of primitivity of the Markov chain.

        Examples:
            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/4, 1/4],
            ...     [1/2,   0, 1/2],
            ...     [1/4, 1/4, 1/2],
            ... ])
            >>> chain.index_of_primitivity()
            2

            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/2,   0],
            ...     [1/2, 1/2,   0],
            ...     [  0,   0,   1],
            ... ])
            >>> chain.index_of_primitivity()
            Traceback (most recent call last):
            ...
            ValueError: chain is not regular
        """
        adjacency_matrix = self.transition_matrix > 0
        Pn = adjacency_matrix.copy()
        max_n = (self.num_states - 1) ** 2 + 1
        for n in range(1, max_n + 1):
            if np.all(Pn):
                return n
            Pn = (Pn @ adjacency_matrix) > 0
        raise ValueError("chain is not regular")

    @cache
    def absorbing_states(self) -> set[int]:
        r"""
        Returns the subset $\mathcal{A} \subseteq \mathcal{S}$ of absorbing states of the Markov chain. A state $i \in \mathcal{S}$ is *absorbing* if $P_{i,i} = 1$.

        Returns:
            The subset $\mathcal{A}$ of absorbing states in the Markov chain.

        Examples:
            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/4, 1/4],
            ...     [1/2,   0, 1/2],
            ...     [1/4, 1/4, 1/2],
            ... ])
            >>> chain.absorbing_states()
            set()

            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/2,   0],
            ...     [1/2,   0, 1/2],
            ...     [  0,   0,   1],
            ... ])
            >>> chain.absorbing_states()
            {2}

            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/2,   0],
            ...     [1/2, 1/2,   0],
            ...     [  0,   0,   1],
            ... ])
            >>> chain.absorbing_states()
            {2}
        """
        return {i for i in range(self.num_states) if self.transition_matrix[i, i] == 1}

    @cache
    def is_absorbing(self) -> bool:
        r"""
        Returns whether the Markov chain is absorbing. A Markov chain is *absorbing* if it has at least one absorbing state and each absorbing state is accessible from at least one transient state.

        Returns:
            `True` if the Markov chain is absorbing, `False` otherwise.

        Examples:
            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/4, 1/4],
            ...     [1/2,   0, 1/2],
            ...     [1/4, 1/4, 1/2],
            ... ])
            >>> chain.is_absorbing()
            False

            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/2,   0],
            ...     [1/2,   0, 1/2],
            ...     [  0,   0,   1],
            ... ])
            >>> chain.is_absorbing()
            True

            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/2,   0],
            ...     [1/2, 1/2,   0],
            ...     [  0,   0,   1],
            ... ])
            >>> chain.is_absorbing()
            False
        """
        absorbing = self.absorbing_states()
        if not absorbing:
            return False
        for i in set(range(self.num_states)) - absorbing:
            if not (absorbing & self.accessible_states_from(i)):
                return False
        return True

    @cache
    def mean_number_of_visits(self) -> Array2D[np.floating]:
        r"""
        Computes the mean number of visits from each transient state to each transient state. This is the expected number of times the chain, starting from transient state $i \in \mathcal{T}$, visits transient state $j \in \mathcal{T}$ before being absorbed.

        Notes:
            - This method only applies to absorbing Markov chains.
            - This corresponds to the *fundamental matrix* $N$ of the Markov chain.


        Returns:
            A $|\mathcal{T}| \times |\mathcal{T}|$-matrix $N$ where entry $N_{i,j}$ is the mean number of visits from $i \in \mathcal{T}$ to $j \in \mathcal{T}$.

        Examples:
            >>> chain = komm.MarkovChain([
            ...     [  1,   0,   0,   0],
            ...     [1/2,   0, 1/2,   0],
            ...     [  0, 1/2,   0, 1/2],
            ...     [  0,   0,   0,   1],
            ... ])
            >>> chain.mean_number_of_visits()
            array([[1.33333333, 0.66666667],
                   [0.66666667, 1.33333333]])
        """
        if not self.is_absorbing():
            raise ValueError("chain is not absorbing")
        transient = list(self.transient_states())
        P = self.transition_matrix
        Q = P[np.ix_(transient, transient)]
        I = np.eye(len(transient))
        N = np.linalg.inv(I - Q)
        return N

    @cache
    def mean_time_to_absorption(self) -> Array1D[np.floating]:
        r"""
        Computes the mean time to absorption from each transient state. This is the expected number of steps until the chain, starting from transient state $i \in \mathcal{T}$, is absorbed.

        Notes:
            - This method only applies to absorbing Markov chains.
            - This is obtained by adding all the entries in the $i$-th row of the fundamental matrix $N$.

        Returns:
            A $|\mathcal{T}|$-vector $\mathbf{t}$ where $\mathbf{t}_i$ is the mean time to absorption from $i \in \mathcal{T}$.

        Examples:
            >>> chain = komm.MarkovChain([
            ...     [  1,   0,   0,   0],
            ...     [1/2,   0, 1/2,   0],
            ...     [  0, 1/2,   0, 1/2],
            ...     [  0,   0,   0,   1],
            ... ])
            >>> chain.mean_time_to_absorption()
            array([2., 2.])
        """
        N = self.mean_number_of_visits()
        return N.sum(axis=1)

    @cache
    def absorption_probabilities(self) -> Array2D[np.floating]:
        r"""
        Computes the absorption probabilities from each transient state to each absorbing state. This is the probability of, starting from transient state $i \in \mathcal{T}$, being absorbed in absorbing state $j \in \mathcal{A}$.

        Notes:
            - This method only applies to absorbing Markov chains.
            - This corresponds to the matrix $B = N R$, where $N$ is the fundamental matrix and $R$ is the submatrix of $P$ row-indexed by $\mathcal{T}$ and column-indexed by $\mathcal{A}$.

        Returns:
            A $|\mathcal{T}| \times |\mathcal{A}|$-matrix $B$ where $B_{i,j}$ is the absorption probability from $i \in \mathcal{T}$ to $j \in \mathcal{A}$.

        Examples:
            >>> chain = komm.MarkovChain([
            ...     [  1,   0,   0,   0],
            ...     [1/2,   0, 1/2,   0],
            ...     [  0, 1/2,   0, 1/2],
            ...     [  0,   0,   0,   1],
            ... ])
            >>> chain.absorption_probabilities()
            array([[0.66666667, 0.33333333],
                   [0.33333333, 0.66666667]])
        """
        N = self.mean_number_of_visits()
        absorbing = list(self.absorbing_states())
        transient = list(self.transient_states())
        P = self.transition_matrix
        R = P[np.ix_(transient, absorbing)]
        B = N @ R
        return B

    @cache
    def period(self, state: int) -> int:
        r"""
        Computes the period of a given state. The *period* of a state $i \in \mathcal{S}$ is the largest integer $d$ such that $(P^n)_{i,i} = 0$ whenever $n$ is not divisible by $d$.

        Parameters:
            state: A state $i \in \mathcal{S}$.

        Returns:
            The period of the state $i$.

        Examples:
            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/4, 1/4],
            ...     [1/2,   0, 1/2],
            ...     [1/4, 1/4, 1/2],
            ... ])
            >>> chain.period(1)
            1

            >>> chain = komm.MarkovChain([
            ...     [0, 1, 0],
            ...     [0, 0, 1],
            ...     [1, 0, 0],
            ... ])
            >>> chain.period(0)
            3
        """
        adjacency_matrix = self.transition_matrix > 0
        Pn = adjacency_matrix.copy()
        periods: list[int] = []
        for n in range(1, self.num_states + 1):
            if Pn[state, state]:
                periods.append(n)
                if reduce(gcd, periods) == 1:
                    return 1
            Pn = (Pn @ adjacency_matrix) > 0
        if not periods:
            return 0
        return reduce(gcd, periods)

    @cache
    def is_aperiodic(self) -> bool:
        r"""
        Returns whether the Markov chain is aperiodic. A Markov chain is *aperiodic* if all states have period $1$.

        Returns:
            `True` if the Markov chain is aperiodic, `False` otherwise.

        Examples:
            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/4, 1/4],
            ...     [1/2,   0, 1/2],
            ...     [1/4, 1/4, 1/2],
            ... ])
            >>> chain.is_aperiodic()
            True

            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/2,   0],
            ...     [1/2, 1/2,   0],
            ...     [  0,   0,   1],
            ... ])
            >>> chain.is_aperiodic()
            True

            >>> chain = komm.MarkovChain([
            ...     [0, 1, 0],
            ...     [0, 0, 1],
            ...     [1, 0, 0],
            ... ])
            >>> chain.is_aperiodic()
            False
        """
        return all(self.period(i) == 1 for i in range(self.num_states))

    def simulate(self, initial_state: int, steps: int) -> Array1D[np.integer]:
        r"""
        Returns random samples from the Markov chain.

        Parameters:
            steps: The number of steps to simulate.

            initial_state: The state $i \in \mathcal{S}$ from which to start the simulation.

        Returns:
            output: A 1D-array of integers in $\mathcal{S}$ representing the states visited during the simulation. The length of the array is equal to `steps`.

        Examples:
            >>> chain = komm.MarkovChain([
            ...     [1/2, 1/4, 1/4],
            ...     [1/2,   0, 1/2],
            ...     [1/4, 1/4, 1/2],
            ... ])
            >>> chain.simulate(initial_state=1, steps=10)
            array([1, 2, 1, 2, 2, 0, 2, 2, 2, 0])

            >>> chain = komm.MarkovChain([
            ...     [0, 1, 0],
            ...     [0, 0, 1],
            ...     [1, 0, 0],
            ... ])
            >>> chain.simulate(initial_state=2, steps=10)
            array([2, 0, 1, 2, 0, 1, 2, 0, 1, 2])
        """
        P = self.transition_matrix
        state = initial_state
        output = np.empty(steps, dtype=int)
        for t in range(steps):
            output[t] = state
            state = self._rng.choice(self.num_states, p=P[state])
        return output

    def simulate_until_absorption(self, initial_state: int) -> Array1D[np.integer]:
        r"""
        Returns random samples from the Markov chain until an absorbing state is reached.

        Note:
            This method only applies to absorbing Markov chains.

        Parameters:
            initial_state: The state $i \in \mathcal{S}$ from which to start the simulation.

        Returns:
            output: A 1D-array of integers in $\mathcal{S}$ representing the states visited during the simulation.

        Examples:
            >>> chain = komm.MarkovChain([
            ...     [  1,   0,   0,   0],
            ...     [1/2,   0, 1/2,   0],
            ...     [  0, 1/2,   0, 1/2],
            ...     [  0,   0,   0,   1],
            ... ])
            >>> chain.simulate_until_absorption(initial_state=1)
            array([1, 2, 1, 2, 3])
        """
        if not self.is_absorbing():
            raise ValueError("chain is not absorbing")
        P = self.transition_matrix
        state = initial_state
        output = [state]
        while state in self.transient_states():
            state = self._rng.choice(self.num_states, p=P[state])
            output.append(state)
        return np.array(output)



from functools import cache
from itertools import combinations

import numpy as np
import numpy.typing as npt
from typeguard import typechecked

from .._util.docs import mkdocstrings
from .SystematicBlockCode import SystematicBlockCode
from .util import extended_parity_submatrix


@mkdocstrings(filters=["!.*"])
@typechecked
class HammingCode(SystematicBlockCode):
    r"""
    Hamming code. For a given parameter $\mu \geq 2$, it is the [linear block code](/ref/BlockCode) with check matrix whose columns are all the $2^\mu - 1$ nonzero binary $\mu$-tuples. The Hamming code has the following parameters:

    - Length: $n = 2^\mu - 1$
    - Dimension: $k = 2^\mu - \mu - 1$
    - Redundancy: $m = \mu$
    - Minimum distance: $d = 3$

    In its extended version, the Hamming code has the following parameters:

    - Length: $n = 2^\mu$
    - Dimension: $k = 2^\mu - \mu - 1$
    - Redundancy: $m = \mu + 1$
    - Minimum distance: $d = 4$

    For more details, see <cite>LC04, Sec. 4.1</cite>.

    Notes:
        - For $\mu = 2$ it reduces to the [repetition code](/ref/RepetitionCode) of length $3$.
        - Its dual is the [simplex code](/ref/SimplexCode).
        - Hamming codes are perfect codes.

    Parameters:
        mu: The parameter $\mu$ of the code. Must satisfy $\mu \geq 2$.
        extended: Whether to use the extended version of the Hamming code. Default is `False`.

    This class represents the code in [systematic form](/ref/SystematicBlockCode), with the information set on the left.

    Examples:
        >>> code = komm.HammingCode(3)
        >>> (code.length, code.dimension, code.redundancy)
        (7, 4, 3)
        >>> code.generator_matrix
        array([[1, 0, 0, 0, 1, 1, 0],
               [0, 1, 0, 0, 1, 0, 1],
               [0, 0, 1, 0, 0, 1, 1],
               [0, 0, 0, 1, 1, 1, 1]])
        >>> code.check_matrix
        array([[1, 1, 0, 1, 1, 0, 0],
               [1, 0, 1, 1, 0, 1, 0],
               [0, 1, 1, 1, 0, 0, 1]])
        >>> code.minimum_distance()
        3

        >>> code = komm.HammingCode(3, extended=True)
        >>> (code.length, code.dimension, code.redundancy)
        (8, 4, 4)
        >>> code.generator_matrix
        array([[1, 0, 0, 0, 1, 1, 0, 1],
               [0, 1, 0, 0, 1, 0, 1, 1],
               [0, 0, 1, 0, 0, 1, 1, 1],
               [0, 0, 0, 1, 1, 1, 1, 0]])
        >>> code.check_matrix
        array([[1, 1, 0, 1, 1, 0, 0, 0],
               [1, 0, 1, 1, 0, 1, 0, 0],
               [0, 1, 1, 1, 0, 0, 1, 0],
               [1, 1, 1, 0, 0, 0, 0, 1]])
        >>> code.minimum_distance()
        4
    """

    def __init__(self, mu: int, extended: bool = False) -> None:
        if not mu >= 2:
            raise ValueError("'mu' must be at least 2")
        self.mu = mu
        self.extended = extended
        super().__init__(parity_submatrix=hamming_parity_submatrix(mu, extended))

    def __repr__(self) -> str:
        args = f"mu={self.mu}, extended={self.extended}"
        return f"{self.__class__.__name__}({args})"

    @cache
    def minimum_distance(self) -> int:
        return 4 if self.extended else 3


def hamming_parity_submatrix(m: int, extended: bool = False) -> npt.NDArray[np.integer]:
    parity_submatrix = np.zeros((2**m - m - 1, m), dtype=int)
    i = 0
    for w in range(2, m + 1):
        for idx in combinations(range(m), w):
            parity_submatrix[i, list(idx)] = 1
            i += 1
    if extended:
        parity_submatrix = extended_parity_submatrix(parity_submatrix)
    return parity_submatrix



from functools import cached_property

import numpy as np
import numpy.typing as npt


class MooreMachine:
    r"""
    Finite-state Moore machine. It is defined by a *set of states* $\mathcal{S}$, an *input alphabet* $\mathcal{X}$, an *output alphabet* $\mathcal{Y}$, a *transition function* $T : \mathcal{S} \times \mathcal{X} \to \mathcal{S}$, and an *output function* $G : \mathcal{S} \to \mathcal{Y}$. Here, for simplicity, the set of states, the input alphabet, and the output alphabet are taken as $\mathcal{S} = [0 : |\mathcal{S}|)$, $\mathcal{X} = [0 : |\mathcal{X}|)$, and $\mathcal{Y} = [0 : |\mathcal{Y}|)$, respectively. more details, see [Wikipedia: Moore machine](https://en.wikipedia.org/wiki/Moore_machine).

    Parameters:
        transitions: The matrix of transitions of the machine, of shape $|\mathcal{S}| \times |\mathcal{X}|$. The element in row $s \in \mathcal{S}$ and column $x \in \mathcal{X}$ should be $T(s, x) \in \mathcal{S}$, that is, the next state of the machine given that the current state is $s$ and the input is $x$.

        outputs: The vector of outputs of the machine, of shape $|\mathcal{S}|$. The element in position $s \in \mathcal{S}$ should be $G(s) \in \mathcal{Y}$, that is, the output of the machine given that the current state is $s$.

    Examples:
        1. Consider the finite-state Moore machine whose state diagram depicted in the figure below.

            <figure markdown>
            ![Finite-state Moore machine example.](/fig/moore.svg)
            </figure>

            It has set of states $\mathcal{S} = \\{ 0, 1, 2, 3 \\}$, input alphabet $\mathcal{X} = \\{ 0, 1 \\}$, output alphabet $\mathcal{Y} = \\{ 0, 1 \\}$. The transition function $T$ and output function $G$ are given by the tables below.

            <div>
            <span>

            | State $s$ | Input $x$ | Transition $T(s, x)$ |
            | :-------: | :-------: | :------------------: |
            | $0$       | $0$       | $0$                  |
            | $0$       | $1$       | $1$                  |
            | $1$       | $0$       | $2$                  |
            | $1$       | $1$       | $3$                  |
            | $2$       | $0$       | $0$                  |
            | $2$       | $1$       | $1$                  |
            | $3$       | $0$       | $2$                  |
            | $3$       | $1$       | $3$                  |

            </span>
            <span>

            | State $s$ |  Output $G(s)$ |
            | :-------: |  :-----------: |
            | $0$       |  $0$           |
            | $1$       |  $0$           |
            | $2$       |  $1$           |
            | $3$       |  $1$           |

            </span>
            </div>

                >>> machine = komm.MooreMachine(
                ...     transitions=[[0, 1], [2, 3], [0, 1], [2, 3]],
                ...     outputs=[0, 0, 1, 1],
                ... )
    """

    def __init__(self, transitions: npt.ArrayLike, outputs: npt.ArrayLike):
        self.transitions = np.asarray(transitions)
        self.outputs = np.asarray(outputs)

    def __repr__(self) -> str:
        args = ", ".join([
            f"transitions={self.transitions.tolist()}",
            f"outputs={self.outputs.tolist()}",
        ])
        return f"{self.__class__.__name__}({args})"

    @cached_property
    def num_states(self) -> int:
        r"""
        The number of states of the machine.

        Examples:
            >>> machine = komm.MooreMachine(
            ...     transitions=[[0, 1], [2, 3], [0, 1], [2, 3]],
            ...     outputs=[0, 0, 1, 1],
            ... )
            >>> machine.num_states
            4
        """
        return self.transitions.shape[0]

    @cached_property
    def num_input_symbols(self) -> int:
        r"""
        The size (cardinality) of the input alphabet $\mathcal{X}$.

        Examples:
            >>> machine = komm.MooreMachine(
            ...     transitions=[[0, 1], [2, 3], [0, 1], [2, 3]],
            ...     outputs=[0, 0, 1, 1],
            ... )
            >>> machine.num_input_symbols
            2
        """
        return self.transitions.shape[1]

    @cached_property
    def num_output_symbols(self) -> int:
        r"""
        The size (cardinality) of the output alphabet $\mathcal{Y}$.

        Examples:
            >>> machine = komm.MooreMachine(
            ...     transitions=[[0, 1], [2, 3], [0, 1], [2, 3]],
            ...     outputs=[0, 0, 1, 1],
            ... )
            >>> machine.num_output_symbols
            2
        """
        return int(np.amax(self.outputs)) + 1

    def process(
        self,
        input: npt.ArrayLike,
        initial_state: int,
    ) -> tuple[npt.NDArray[np.integer], int]:
        r"""
        Returns the output sequence corresponding to a given input sequence. It assumes the machine starts at a given initial state $s_\mathrm{i}$. The input sequence and the output sequence are denoted by $x = (x_0, x_1, \ldots, x_{L-1}) \in \mathcal{X}^L$ and $y = (y_0, y_1, \ldots, y_{L-1}) \in \mathcal{Y}^{L}$, respectively.

        Parameters:
            input: The input sequence $x \in \mathcal{X}^L$. It should be a 1D-array with elements in $\mathcal{X}$.

            initial_state: The initial state $s_\mathrm{i}$ of the machine. Should be an integer in $\mathcal{S}$.

        Returns:
            output: The output sequence $y \in \mathcal{Y}^{L}$ corresponding to `input`, assuming the machine starts at the state given by `initial_state`. It is a 1D-array with elements in $\mathcal{Y}$.

            final_state: The final state $s_\mathrm{f}$ of the machine. It is an integer in $\mathcal{S}$.

        Examples:
            >>> machine = komm.MooreMachine(
            ...     transitions=[[0, 1], [2, 3], [0, 1], [2, 3]],
            ...     outputs=[0, 0, 1, 1],
            ... )
            >>> input, initial_state = [1, 1, 0, 1, 0], 0
            >>> output, final_state = machine.process(input, initial_state)
            >>> output
            array([0, 1, 1, 0, 1])
            >>> final_state
            2
        """
        output = np.empty_like(input, dtype=int)
        s = initial_state
        for t, x in np.ndenumerate(input):
            s = self.transitions[s, x]
            output[t] = self.outputs[s]
        final_state = int(s)
        return output, final_state



from functools import cache, reduce
from operator import itemgetter

import numpy as np
import numpy.typing as npt
from tqdm import tqdm

from .._util.docs import mkdocstrings
from .._util.validators import validate_pmf
from ..types import Array1D
from .FixedToVariableCode import FixedToVariableCode
from .util import Word, lexicographical_code


@mkdocstrings(filters=["!.*"])
class FanoCode(FixedToVariableCode):
    r"""
    Binary Fano code. For a given pmf $p$ over $\mathcal{X}$, it is a [fixed-to-variable length code](/ref/FixedToVariableCode) in which the source words are first sorted in descending order of probability and then are recursively partitioned into two groups of approximately equal total probability, assigning bit $\mathtt{0}$ to one group and bit $\mathtt{1}$ to the other, until each source word is assigned a unique codeword. For more details, see [Wikipedia: Shannon–Fano coding](https://en.wikipedia.org/wiki/Shannon%E2%80%93Fano_coding).

    Notes:
        Fano codes are always [prefix-free](/ref/FixedToVariableCode/#is_prefix_free) (hence [uniquely decodable](/ref/FixedToVariableCode/#is_uniquely_decodable)).

    Parameters:
        pmf: The pmf $p$ to be considered. It must be a one-dimensional array of floats of size $|\mathcal{X}|$. The elements must be non-negative and sum to $1$.

        source_block_size: The source block size $k$. The default value is $k = 1$.

    Examples:
        >>> pmf = [0.8, 0.1, 0.1]

        >>> code = komm.FanoCode(pmf)
        >>> code.enc_mapping
        {(0,): (0,),
         (1,): (1, 0),
         (2,): (1, 1)}
        >>> code.rate(pmf)  # doctest: +FLOAT_CMP
        np.float64(1.2)

        >>> code = komm.FanoCode(pmf, 2)
        >>> code.enc_mapping
        {(0, 0): (0,),
         (0, 1): (1, 0, 0),
         (0, 2): (1, 0, 1),
         (1, 0): (1, 1, 0),
         (1, 1): (1, 1, 1, 1, 0, 0),
         (1, 2): (1, 1, 1, 1, 0, 1),
         (2, 0): (1, 1, 1, 0),
         (2, 1): (1, 1, 1, 1, 1, 0),
         (2, 2): (1, 1, 1, 1, 1, 1)}
        >>> code.rate(pmf)  # doctest: +FLOAT_CMP
        np.float64(0.96)
    """

    def __init__(self, pmf: npt.ArrayLike, source_block_size: int = 1):
        self.pmf = validate_pmf(pmf)
        if not source_block_size >= 1:
            raise ValueError("'source_block_size' must be at least 1")
        super().__init__(
            source_cardinality=self.pmf.size,
            target_cardinality=2,
            source_block_size=source_block_size,
            enc_mapping=fano_code(self.pmf, source_block_size),
        )

    def __repr__(self) -> str:
        args = ", ".join([
            f"pmf={self.pmf.tolist()}",
            f"source_block_size={self.source_block_size}",
        ])
        return f"{self.__class__.__name__}({args})"

    @cache
    def is_uniquely_decodable(self) -> bool:
        return True

    @cache
    def is_prefix_free(self) -> bool:
        return True


def fano_code_lengths(pmf: Array1D[np.floating]) -> Array1D[np.integer]:
    pbar = tqdm(desc="Generating Fano code lengths", total=pmf.size, delay=2.5)
    lengths = np.zeros_like(pmf, dtype=int)
    items = sorted(np.ndenumerate(pmf), key=itemgetter(1), reverse=True)
    stack: list[tuple[int, int, int]] = [(0, pmf.size, 0)]
    while stack:
        lo, hi, length = stack.pop()
        if hi - lo == 1:
            x, _ = items[lo]
            lengths[x] = length
            pbar.update()
            continue
        probs = [p for _, p in items[lo:hi]]
        index = int(np.argmin(np.abs(np.cumsum(probs) - np.sum(probs) / 2)))
        mid = lo + index + 1
        stack.append((mid, hi, length + 1))
        stack.append((lo, mid, length + 1))
    pbar.close()
    return lengths


def fano_code(pmf: Array1D[np.floating], source_block_size: int) -> dict[Word, Word]:
    extended_pmf = reduce(np.multiply.outer, [pmf] * source_block_size)
    lengths = fano_code_lengths(extended_pmf.ravel())
    codewords = lexicographical_code(lengths)
    enc_mapping: dict[Word, Word] = {}
    for x, c in zip(np.ndindex(extended_pmf.shape), codewords):
        enc_mapping[x] = c
    return enc_mapping



from dataclasses import dataclass
from typing import Literal

import numpy as np
import numpy.typing as npt

from .. import abc
from .._util.decorators import blockwise


@dataclass
class ExhaustiveSearchDecoder(abc.BlockDecoder[abc.BlockCode]):
    r"""
    Exhaustive search decoder for general [block codes](/ref/BlockCode). This decoder implements a brute-force search over all possible codewords to find the one that is closest (in terms of Hamming distance, for hard-decision decoding, or Euclidean distance, for soft-decision decoding) to the received word.

    Parameters:
        code: The block code to be used for decoding.
        input_type: The type of the input. Either `'hard'` or `'soft'`. Default is `'hard'`.

    Notes:
        - Input type: `hard` (bits) or `soft` (either L-values or channel outputs).
        - Output type: `hard` (bits).
    """

    code: abc.BlockCode
    input_type: Literal["hard", "soft"] = "hard"

    def __post_init__(self) -> None:
        self._codewords = self.code.codewords()

    def decode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer | np.floating]:
        r"""
        Examples:
            >>> code = komm.HammingCode(3)

            >>> decoder = komm.ExhaustiveSearchDecoder(code, input_type="hard")
            >>> decoder.decode([[1, 1, 0, 1, 0, 1, 1], [1, 0, 1, 1, 0, 0, 0]])
            array([[1, 1, 0, 0],
                   [1, 0, 1, 1]])

            >>> decoder = komm.ExhaustiveSearchDecoder(code, input_type="soft")
            >>> decoder.decode([-1.3, -0.8, +1.1, -0.8, +1.2, -0.2, -1.4])
            array([1, 1, 0, 0])
        """

        @blockwise(self.code.length)
        def decode(r: npt.NDArray[np.integer]):
            if self.input_type == "hard":
                ds = r[..., np.newaxis, :] != self._codewords
            else:
                ds = -r[..., np.newaxis, :] * (-1) ** self._codewords
            metrics = np.sum(ds, axis=-1)
            v_hat = self._codewords[np.argmin(metrics, axis=-1)]
            u_hat = self.code.project_word(v_hat)
            return u_hat

        return decode(input)



from functools import cache, reduce

import numpy as np
import numpy.typing as npt

from .._util.docs import mkdocstrings
from .._util.validators import validate_pmf
from ..types import Array1D
from .FixedToVariableCode import FixedToVariableCode
from .util import Word, lexicographical_code


@mkdocstrings(filters=["!.*"])
class ShannonCode(FixedToVariableCode):
    r"""
    Binary Shannon code. For a given pmf $p$ over $\mathcal{X}$, it is a [fixed-to-variable length code](/ref/FixedToVariableCode) in which the length of the codeword $\Enc(\mathbf{x})$ associated with a source word $\mathbf{x} \in \mathcal{X}^k$ is given by
    $$
        \ell(\mathbf{x}) = \left\lceil \log_2 \frac{1}{p(\mathbf{x})} \right\rceil.
    $$
    This function implements the lexicographic order assignment as described in [Wikipedia: Shannon–Fano coding](https://en.wikipedia.org/wiki/Shannon%E2%80%93Fano_coding).

    Notes:
        Shannon codes are always [prefix-free](/ref/FixedToVariableCode/#is_prefix_free) (hence [uniquely decodable](/ref/FixedToVariableCode/#is_uniquely_decodable)).

    Parameters:
        pmf: The pmf $p$ to be considered. It must be a one-dimensional array of floats of size $|\mathcal{X}|$. The elements must be non-negative and sum to $1$.

        source_block_size: The source block size $k$. The default value is $k = 1$.

    Examples:
        >>> pmf = [0.8, 0.1, 0.1]

        >>> code = komm.ShannonCode(pmf)
        >>> code.enc_mapping
        {(0,): (0,),
         (1,): (1, 0, 0, 0),
         (2,): (1, 0, 0, 1)}
        >>> code.rate(pmf)  # doctest: +FLOAT_CMP
        np.float64(1.6)

        >>> code = komm.ShannonCode(pmf, 2)
        >>> code.enc_mapping
        {(0, 0): (0,),
         (0, 1): (1, 0, 0, 0),
         (0, 2): (1, 0, 0, 1),
         (1, 0): (1, 0, 1, 0),
         (1, 1): (1, 1, 0, 0, 0, 0, 0),
         (1, 2): (1, 1, 0, 0, 0, 0, 1),
         (2, 0): (1, 0, 1, 1),
         (2, 1): (1, 1, 0, 0, 0, 1, 0),
         (2, 2): (1, 1, 0, 0, 0, 1, 1)}
        >>> code.rate(pmf)  # doctest: +FLOAT_CMP
        np.float64(1.1)
    """

    def __init__(self, pmf: npt.ArrayLike, source_block_size: int = 1):
        self.pmf = validate_pmf(pmf)
        if not source_block_size >= 1:
            raise ValueError("'source_block_size' must be at least 1")
        super().__init__(
            source_cardinality=self.pmf.size,
            target_cardinality=2,
            source_block_size=source_block_size,
            enc_mapping=shannon_code(self.pmf, source_block_size),
        )

    def __repr__(self) -> str:
        args = ", ".join([
            f"pmf={self.pmf.tolist()}",
            f"source_block_size={self.source_block_size}",
        ])
        return f"{self.__class__.__name__}({args})"

    @cache
    def is_uniquely_decodable(self) -> bool:
        return True

    @cache
    def is_prefix_free(self) -> bool:
        return True


def shannon_code_lengths(pmf: Array1D[np.floating]) -> Array1D[np.integer]:
    lengths = np.zeros_like(pmf, dtype=int)
    mask = pmf > 0
    if np.sum(pmf**2) == 1:  # Deterministic case
        lengths[mask] = 1
    else:
        lengths[mask] = np.ceil(np.log2(1 / pmf[mask])).astype(int)
    return lengths


def shannon_code(pmf: Array1D[np.floating], source_block_size: int) -> dict[Word, Word]:
    extended_pmf = reduce(np.multiply.outer, [pmf] * source_block_size)
    lengths = shannon_code_lengths(extended_pmf.ravel())
    codewords = lexicographical_code(lengths)
    enc_mapping: dict[Word, Word] = {}
    for x, c in zip(np.ndindex(extended_pmf.shape), codewords):
        enc_mapping[x] = c
    return enc_mapping



from math import ceil, log

import numpy as np
import numpy.typing as npt
from tqdm import tqdm

from .._util.validators import validate_integer_range
from .util import integer_to_symbols, symbols_to_integer

Token = tuple[int, int, int]


class LempelZiv77Code:
    r"""
    Lempel–Ziv 77 (LZ77 or LZ1) code. It is a lossless data compression algorithm which is asymptotically optimal for ergodic sources. Let $\mathcal{X}$ be the source alphabet, and $\mathcal{Y}$ be the target alphabet. The notation used here is the following: $S \geq 1$ is the size of the *search buffer*, $L \geq 1$ is the size of the *lookahead buffer*, and $W = S + L$ is the size of the *sliding window*. The token format follows the [original LZ77 paper](https://doi.org/10.1109%2FTIT.1977.1055714), namely $(p, \ell, x)$, where $p \in [0 : S)$ is the *pointer* for the match, $\ell \in [0 : L)$ is the *length* of the match, and $x \in \mathcal{X}$ is the source symbol following the match, but with both $p$ and $\ell$ being $0$-indexed instead of $1$-indexed. Also following the LZ77 original paper, a token is represented as a fixed-size word in $\mathcal{Y}^n$, where $$n = \log S + \log L + \log |\mathcal{X}|$$ and all logs are to base $|\mathcal{Y}|$. For more details, see <cite>Say06, Sec. 5.4.1</cite> and <cite>CT06, Sec. 13.4.1</cite>.

    Note:
        Here, for simplicity, we assume that the source alphabet is $\mathcal{X} = [0 : |\mathcal{X}|)$ and the target alphabet is $\mathcal{Y} = [0 : |\mathcal{Y}|)$, where $|\mathcal{X}| \geq 2$ and $|\mathcal{Y}| \geq 2$ are called the *source cardinality* and *target cardinality*, respectively.

    Parameters:
        search_size: The search buffer size $S$. Must satisfy $S \geq 1$.
        lookahead_size: The lookahead buffer size $L$. Must satisfy $L \geq 1$.
        source_cardinality: The source cardinality $|\mathcal{X}|$. Must satisfy $|\mathcal{X}| \geq 2$.
        target_cardinality: The target cardinality $|\mathcal{Y}|$. Must satisfy $|\mathcal{Y}| \geq 2$. The default value is $2$ (binary).
        search_buffer: The initial state of the search buffer. Must be a 1D-array of length $S$ with elements in $\mathcal{X}$. The default value corresponds to $(0, \ldots, 0) \in \mathcal{X}^S$.

    Examples:
        >>> lz77 = komm.LempelZiv77Code(
        ...     search_size=2**13,
        ...     lookahead_size=16,
        ...     source_cardinality=256,
        ...     target_cardinality=2,
        ... )
    """

    def __init__(
        self,
        *,
        search_size: int,
        lookahead_size: int,
        source_cardinality: int,
        target_cardinality: int = 2,
        search_buffer: npt.ArrayLike | None = None,
    ):
        if not search_size >= 1:
            raise ValueError("'search_size' must be at least 1")
        if not lookahead_size >= 1:
            raise ValueError("'lookahead_size' must be at least 1")
        if not source_cardinality >= 2:
            raise ValueError("'source_cardinality' must be at least 2")
        if not target_cardinality >= 2:
            raise ValueError("'target_cardinality' must be at least 2")

        self.search_size = search_size
        self.lookahead_size = lookahead_size
        self.source_cardinality = source_cardinality
        self.target_cardinality = target_cardinality

        if search_buffer is None:
            self.search_buffer = [0] * self.search_size
        else:
            self.search_buffer = np.asarray(search_buffer, dtype=int).tolist()
            if len(self.search_buffer) != self.search_size:
                raise ValueError(
                    "length of 'search_buffer' must be S ="
                    f" {self.search_size} (got {len(self.search_buffer)})"
                )

    def __repr__(self) -> str:
        args = ", ".join([
            f"search_size={self.search_size}",
            f"lookahead_size={self.lookahead_size}",
            f"source_cardinality={self.source_cardinality}",
            f"target_cardinality={self.target_cardinality}",
        ])
        return f"{self.__class__.__name__}({args})"

    @property
    def window_size(self) -> int:
        r"""
        The sliding window size $W$. It is given by $W = S + L$.
        """
        return self.search_size + self.lookahead_size

    def _get_widths(self) -> tuple[int, int, int]:
        calY = self.target_cardinality
        p_width = ceil(log(self.search_size, calY))
        l_width = ceil(log(self.lookahead_size, calY))
        x_width = ceil(log(self.source_cardinality, calY))
        return p_width, l_width, x_width

    def source_to_tokens(self, source: npt.ArrayLike) -> list[Token]:
        r"""
        Encodes a given sequence of source symbols to the corresponding list of tokens.

        Examples:
            >>> lz77 = komm.LempelZiv77Code(
            ...     search_size=9,
            ...     lookahead_size=9,
            ...     source_cardinality=3,
            ...     target_cardinality=3,
            ... )
            >>> lz77.source_to_tokens([0, 0, 1, 0, 1, 0, 2, 1, 0, 2, 1, 0, 2, 1, 2])
            [(8, 2, 1), (7, 3, 2), (6, 7, 2)]
        """
        source = validate_integer_range(source, high=self.source_cardinality)
        ss, ls = self.search_size, self.lookahead_size
        buffer = bytes(self.search_buffer + source.tolist())
        tokens: list[Token] = []

        pbar = tqdm(total=len(buffer), initial=ss, desc="Compressing LZ77", delay=2.5)
        i = ss
        while i < len(buffer):
            l = min(ls, len(buffer) - i) - 1
            while True:
                p = buffer[i - ss : i + l - 1].rfind(buffer[i : i + l])
                if p >= 0:
                    break
                l -= 1
            tokens.append((p, l, buffer[i + l]))
            i += l + 1
            pbar.update(l + 1)

        return tokens

    def tokens_to_source(self, tokens: list[Token]) -> npt.NDArray[np.integer]:
        r"""
        Decodes a given list of tokens to the corresponding sequence of source symbols.

        Examples:
            >>> lz77 = komm.LempelZiv77Code(
            ...     search_size=9,
            ...     lookahead_size=9,
            ...     source_cardinality=3,
            ...     target_cardinality=3,
            ... )
            >>> lz77.tokens_to_source([(8, 2, 1), (7, 3, 2), (6, 7, 2)])
            array([0, 0, 1, 0, 1, 0, 2, 1, 0, 2, 1, 0, 2, 1, 2])
        """
        ss = self.search_size
        buffer = self.search_buffer.copy()
        for p, l, x in tokens:
            start = len(buffer) - ss + p
            for j in range(l):
                buffer.append(buffer[start + j])
            buffer.append(x)
        source = np.array(buffer[ss:], dtype=int)
        return source

    def tokens_to_target(self, tokens: list[Token]) -> npt.NDArray[np.integer]:
        r"""
        Returns the target alphabet representation corresponding to a given list of tokens.

        Examples:
            >>> lz77 = komm.LempelZiv77Code(
            ...     search_size=9,
            ...     lookahead_size=9,
            ...     source_cardinality=3,
            ...     target_cardinality=3,
            ... )
            >>> lz77.tokens_to_target([(8, 2, 1), (7, 3, 2), (6, 7, 2)])
            array([2, 2, 0, 2, 1, 2, 1, 1, 0, 2, 2, 0, 2, 1, 2])
        """
        calY = self.target_cardinality
        p_width, l_width, x_width = self._get_widths()
        target: list[int] = []
        for p, l, x in tokens:
            target.extend(integer_to_symbols(p, base=calY, width=p_width))
            target.extend(integer_to_symbols(l, base=calY, width=l_width))
            target.extend(integer_to_symbols(x, base=calY, width=x_width))
        return np.array(target, dtype=int)

    def target_to_tokens(self, target: npt.ArrayLike) -> list[Token]:
        r"""
        Returns the list of tokens corresponding to a given target alphabet representation.

        Examples:
            >>> lz77 = komm.LempelZiv77Code(
            ...     search_size=9,
            ...     lookahead_size=9,
            ...     source_cardinality=3,
            ...     target_cardinality=3,
            ... )
            >>> lz77.target_to_tokens([2, 2, 0, 2, 1, 2, 1, 1, 0, 2, 2, 0, 2, 1, 2])
            [(8, 2, 1), (7, 3, 2), (6, 7, 2)]
        """
        target = np.asarray(target, dtype=int)
        calY = self.target_cardinality
        p_width, l_width, x_width = self._get_widths()
        tokens: list[Token] = []
        i = 0
        while i + p_width + l_width + x_width <= target.size:
            p = int(symbols_to_integer(target[i : i + p_width], base=calY))
            i += p_width
            l = int(symbols_to_integer(target[i : i + l_width], base=calY))
            i += l_width
            x = int(symbols_to_integer(target[i : i + x_width], base=calY))
            i += x_width
            tokens.append((p, l, x))
        return tokens

    def encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Encodes a sequence of source symbols to a sequence of target symbols.

        Parameters:
            input: The sequence of source symbols to be encoded. Must be a 1D-array with elements in $\mathcal{X}$.

        Returns:
            output: The sequence of encoded target symbols. It is a 1D-array with elements in $\mathcal{Y}$.

        Examples:
            >>> lz77 = komm.LempelZiv77Code(
            ...     search_size=9,
            ...     lookahead_size=9,
            ...     source_cardinality=3,
            ...     target_cardinality=3,
            ... )
            >>> lz77.encode([0, 0, 1, 0, 1, 0, 2, 1, 0, 2, 1, 0, 2, 1, 2])
            array([2, 2, 0, 2, 1, 2, 1, 1, 0, 2, 2, 0, 2, 1, 2])
        """
        tokens = self.source_to_tokens(input)
        output = self.tokens_to_target(tokens)
        return output

    def decode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Decodes a sequence of target symbols to a sequence of source symbols.

        Parameters:
            input: The sequence of target symbols to be decoded. Must be a 1D-array with elements in $\mathcal{Y}$. Also, the sequence must be a valid output of the `encode` method.

        Returns:
            output: The sequence of decoded source symbols. It is a 1D-array with elements in $\mathcal{X}$.

        Examples:
            >>> lz77 = komm.LempelZiv77Code(
            ...     search_size=9,
            ...     lookahead_size=9,
            ...     source_cardinality=3,
            ...     target_cardinality=3,
            ... )
            >>> lz77.decode([2, 2, 0, 2, 1, 2, 1, 1, 0, 2, 2, 0, 2, 1, 2])
            array([0, 0, 1, 0, 1, 0, 2, 1, 0, 2, 1, 0, 2, 1, 2])
        """
        tokens = self.target_to_tokens(input)
        output = self.tokens_to_source(tokens)
        return output



from functools import cache, cached_property
from itertools import count, product

import numpy as np
import numpy.typing as npt
from typing_extensions import Self

from .._util.validators import validate_pmf
from ..types import Array1D
from .util import (
    Word,
    is_prefix_free,
    is_uniquely_parsable,
    parse_fixed_length,
    parse_prefix_free,
)


class FixedToVariableCode:
    r"""
    General fixed-to-variable length code. A *fixed-to-variable length code* with *source alphabet* $\mathcal{X}$, *target alphabet* $\mathcal{Y}$, and *source block size* $k$ is defined by an *encoding mapping* $\Enc: \mathcal{X}^k \to \mathcal{Y}^+$, where the domain is the set of all $k$-tuples with entries in $\mathcal{X}$, and the co-domain is the set of all finite-length, non-empty tuples with entries in $\mathcal{Y}$. The elements in the image of $\Enc$ are called *codewords*.

    Note:
        Here, for simplicity, we assume that the source alphabet is $\mathcal{X} = [0 : |\mathcal{X}|)$ and the target alphabet is $\mathcal{Y} = [0 : |\mathcal{Y}|)$, where $|\mathcal{X}| \geq 2$ and $|\mathcal{Y}| \geq 2$ are called the *source cardinality* and *target cardinality*, respectively.
    """

    def __init__(
        self,
        source_cardinality: int,
        target_cardinality: int,
        source_block_size: int,
        enc_mapping: dict[Word, Word],
    ) -> None:
        self._source_cardinality = source_cardinality
        self._target_cardinality = target_cardinality
        self._source_block_size = source_block_size
        self._enc_mapping = enc_mapping
        self.__post_init__()

    def __post_init__(self) -> None:
        domain, codomain = self.enc_mapping.keys(), self.enc_mapping.values()
        calX, calY = self.source_cardinality, self.target_cardinality
        k = self.source_block_size
        if not calX >= 2:
            raise ValueError("'source_cardinality' must be at least 2")
        if not calY >= 2:
            raise ValueError("'target_cardinality' must be at least 2")
        if not k >= 1:
            raise ValueError("'source_block_size' must be at least 1")
        if set(domain) != set(product(range(calX), repeat=k)):
            raise ValueError("'enc_mapping': invalid domain")
        if not all(all(0 <= x < calY for x in word) for word in codomain):
            raise ValueError("'enc_mapping': invalid co-domain")

    def __repr__(self) -> str:
        args = ", ".join([
            f"source_cardinality={self.source_cardinality}",
            f"target_cardinality={self.target_cardinality}",
            f"source_block_size={self.source_block_size}",
            f"enc_mapping={self.enc_mapping}",
        ])
        return f"{self.__class__.__name__}({args})"

    @classmethod
    def from_enc_mapping(cls, enc_mapping: dict[Word, Word]) -> Self:
        r"""
        Constructs a fixed-to-variable length code from the encoding mapping $\Enc$.

        Parameters:
            enc_mapping: The encoding mapping $\Enc$. Must be a dictionary whose keys are all the $k$-tuples of integers in $\mathcal{X}$ and whose values are non-empty tuples of integers in $\mathcal{Y}$.

        Notes:
            The source block size $k$ is inferred from the domain of the encoding mapping, and the source and target cardinalities $|\mathcal{X}|$ and $|\mathcal{Y}|$ are inferred from the maximum values in the domain and co-domain, respectively.

        Examples:
            >>> code = komm.FixedToVariableCode.from_enc_mapping({
            ...     (0,): (0,),
            ...     (1,): (1, 0),
            ...     (2,): (1, 1),
            ... })
            >>> code.source_cardinality, code.target_cardinality, code.source_block_size
            (3, 2, 1)
            >>> code.codewords
            [(0,), (1, 0), (1, 1)]

            >>> code = komm.FixedToVariableCode.from_enc_mapping({
            ...     (0, 0): (0,),
            ...     (0, 1): (1, 1),
            ...     (1, 0): (1, 1, 0),
            ...     (1, 1): (1, 0, 1),
            ... })
            >>> code.source_cardinality, code.target_cardinality, code.source_block_size
            (2, 2, 2)
            >>> code.codewords
            [(0,), (1, 1), (1, 1, 0), (1, 0, 1)]
        """
        domain, codomain = enc_mapping.keys(), enc_mapping.values()
        calX = max(max(word) for word in domain) + 1
        calY = max(max(word) for word in codomain) + 1
        k = len(next(iter(domain)))
        return cls(calX, calY, k, enc_mapping)

    @classmethod
    def from_codewords(cls, source_cardinality: int, codewords: list[Word]) -> Self:
        r"""
        Constructs a fixed-to-variable length code from the source cardinality $|\mathcal{X}|$ and a list of codewords.

        Parameters:
            source_cardinality: The source cardinality $|\mathcal{X}|$. Must be an integer greater than or equal to $2$.

            codewords: The codewords of the code. Must be a list of length $|\mathcal{X}|^k$ containing tuples of integers in $\mathcal{Y}$. The tuple in position $i$ must be equal to $\Enc(\mathbf{x})$, where $\mathbf{x}$ is the $i$-th element in the lexicographic ordering of $\mathcal{X}^k$.

        Notes:
            The source block size $k$ is inferred from the length of the codewords, and the target cardinality $|\mathcal{Y}|$ is inferred from the maximum value in the codewords.

        Examples:
            >>> code = komm.FixedToVariableCode.from_codewords(
            ...     source_cardinality=3,
            ...     codewords=[(0,), (1, 0), (1, 1)],
            ... )
            >>> code.source_cardinality, code.target_cardinality, code.source_block_size
            (3, 2, 1)
            >>> code.enc_mapping
            {(0,): (0,),
             (1,): (1, 0),
             (2,): (1, 1)}

            >>> code = komm.FixedToVariableCode.from_codewords(
            ...     source_cardinality=2,
            ...     codewords=[(0,), (1, 1), (1, 1, 0), (1, 0, 1)]
            ... )
            >>> code.source_cardinality, code.target_cardinality, code.source_block_size
            (2, 2, 2)
            >>> code.enc_mapping
            {(0, 0): (0,),
             (0, 1): (1, 1),
             (1, 0): (1, 1, 0),
             (1, 1): (1, 0, 1)}
        """
        calX = source_cardinality
        calY = max(max(codeword) for codeword in codewords) + 1
        k = next(k for k in count(1) if calX**k >= len(codewords))
        enc_mapping = dict(zip(product(range(calX), repeat=k), codewords))
        return cls(calX, calY, k, enc_mapping)

    @cached_property
    def source_cardinality(self) -> int:
        r"""
        The source cardinality $|\mathcal{X}|$ of the code. It is the number of symbols in the source alphabet.

        Examples:
            >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (1, 0), (1, 1)])
            >>> code.source_cardinality
            3
        """
        return self._source_cardinality

    @cached_property
    def target_cardinality(self) -> int:
        r"""
        The target cardinality $|\mathcal{Y}|$ of the code. It is the number of symbols in the target alphabet.

        Examples:
            >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (1, 0), (1, 1)])
            >>> code.target_cardinality
            2
        """
        return self._target_cardinality

    @cached_property
    def source_block_size(self) -> int:
        r"""
        The source block size $k$ of the code. It is the number of symbols in each source word.

        Examples:
            >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (1, 0), (1, 1)])
            >>> code.source_block_size
            1
        """
        return self._source_block_size

    @cached_property
    def size(self) -> int:
        r"""
        The number of codewords in the code. It is equal to $|\mathcal{X}|^k$.

        Examples:
            >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (1, 0), (1, 1)])
            >>> code.size
            3
        """
        return len(self.enc_mapping)

    @cached_property
    def enc_mapping(self) -> dict[Word, Word]:
        r"""
        The encoding mapping $\Enc$ of the code. It is a dictionary of length $|\mathcal{X}|^k$ whose keys are all the $k$-tuples of integers in $\mathcal{X}$ and whose values are the corresponding codewords.

        Examples:
            >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (1, 0), (1, 1)])
            >>> code.enc_mapping
            {(0,): (0,),
             (1,): (1, 0),
             (2,): (1, 1)}
        """
        return self._enc_mapping

    @cached_property
    def _inv_enc_mapping(self) -> dict[Word, Word]:
        return {y: x for x, y in self.enc_mapping.items()}

    @cached_property
    def codewords(self) -> list[Word]:
        r"""
        The codewords of the code. They correspond to the image of the encoding mapping $\Enc$.

        Examples:
            >>> code = komm.FixedToVariableCode.from_enc_mapping({
            ...     (0,): (0,),
            ...     (1,): (1, 0),
            ...     (2,): (1, 1),
            ... })
            >>> code.codewords
            [(0,), (1, 0), (1, 1)]
        """
        return list(self.enc_mapping.values())

    @cached_property
    def _codewords_lengths(self) -> npt.NDArray[np.integer]:
        calX, k = self.source_cardinality, self.source_block_size
        lengths = np.empty((calX,) * k, dtype=int)
        for x, y in self.enc_mapping.items():
            lengths[x] = len(y)
        return lengths

    @cache
    def is_uniquely_decodable(self) -> bool:
        r"""
        Returns whether the code is uniquely decodable. A code is *uniquely decodable* if there is a unique way to parse any concatenation of codewords.

        Examples:
            >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (1, 0), (1, 1)])
            >>> code.is_uniquely_decodable()
            True

            >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (0, 1), (1, 1)])
            >>> code.is_uniquely_decodable()
            True

            >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (0, 1), (1, 0)])
            >>> code.is_uniquely_decodable()  # 010 can be parsed as 0|10 or 01|0
            False
        """
        return is_uniquely_parsable(self.codewords)

    @cache
    def is_prefix_free(self) -> bool:
        r"""
        Returns whether the code is prefix-free. A code is *prefix-free* if no codeword is a prefix of any other codeword.

        Examples:
            >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (1, 0), (1, 1)])
            >>> code.is_prefix_free()
            True

            >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (0, 1), (1, 1)])
            >>> code.is_prefix_free()
            False

            >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (0, 1), (1, 0)])
            >>> code.is_prefix_free()
            False
        """
        return is_prefix_free(self.codewords)

    @cache
    def kraft_parameter(self) -> float:
        r"""
        Computes the Kraft parameter $K$ of the code. This quantity is given by
        $$
            K = \sum_{\mathbf{x} \in \mathcal{X}^k} |\mathcal{Y}|^{-{\ell(\mathbf{x})}},
        $$
        where $\ell(\mathbf{x})$ is the length of the codeword $\Enc(\mathbf{x})$ associated with the source word $\mathbf{x} \in \mathcal{X}^k$.

        Returns:
            kraft_parameter: The Kraft parameter $K$ of the code.

        Examples:
            >>> code = komm.FixedToVariableCode.from_codewords(
            ...     source_cardinality=5,
            ...     codewords=[(0, 0, 0), (0, 0, 1), (0, 1, 0), (1, 0, 1), (1, 1)],
            ... )
            >>> code.kraft_parameter()
            np.float64(0.75)

            >>> code = komm.FixedToVariableCode.from_codewords(
            ...     source_cardinality=4,
            ...     codewords=[(0,), (1, 0), (1, 1, 0), (1, 1, 1)],
            ... )
            >>> code.kraft_parameter()
            np.float64(1.0)

            >>> code = komm.FixedToVariableCode.from_codewords(
            ...     source_cardinality=4,
            ...     codewords=[(0, 0), (1, 1), (0,), (1,)],
            ... )
            >>> code.kraft_parameter()
            np.float64(1.5)
        """
        calY = self.target_cardinality
        lengths = self._codewords_lengths
        return np.sum(np.float_power(calY, -lengths))

    def rate(self, pmf: npt.ArrayLike) -> float:
        r"""
        Computes the expected rate $R$ of the code, considering a given (first-order) pmf $p$ over $\mathcal{X}$. This quantity is given by
        $$
            R = \frac{\bar{n}}{k},
        $$
        where $\bar{n}$ is the expected codeword length, assuming iid source symbols drawn according to $p$. It is measured in $|\mathcal{Y}|$-ary digits per source symbol.

        Parameters:
            pmf: The pmf $p$ to be considered. It must be a one-dimensional array of floats of size $|\mathcal{X}|$. The elements must be non-negative and sum to $1$.

        Returns:
            rate: The expected rate $R$ of the code.

        Examples:
            >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (1, 0), (1, 1)])
            >>> code.rate([1/2, 1/4, 1/4])
            np.float64(1.5)
        """
        k = self.source_block_size
        pmf = validate_pmf(pmf)
        probabilities = [np.prod(ps) for ps in product(pmf, repeat=k)]
        lengths = [len(y) for y in self.codewords]
        return np.dot(lengths, probabilities) / k

    def encode(self, input: npt.ArrayLike) -> Array1D[np.integer]:
        r"""
        Encodes a sequence of source symbols using the code, which must be uniquely decodable.

        Parameters:
            input: The sequence of symbols to be encoded. Must be a 1D-array with elements in $\mathcal{X}$ and have a length that is a multiple of $k$.

        Returns:
            output: The sequence of encoded symbols. It is a 1D-array with elements in $\mathcal{Y}$.

        Examples:
            >>> code = komm.FixedToVariableCode.from_enc_mapping({
            ...     (0, 0): (0,),
            ...     (0, 1): (1, 1),
            ...     (1, 0): (1, 0, 0),
            ...     (1, 1): (1, 0, 1),
            ... })

            >>> code.encode([0, 1, 0, 0])
            array([1, 1, 0])

            >>> code.encode([0, 1, 0])  # Not a multiple of the source block size
            Traceback (most recent call last):
            ...
            ValueError: length of input must be a multiple of block size 2 (got 3)

            >>> code.encode([0, 7, 0, 0])  # 07 is not a valid source word
            Traceback (most recent call last):
            ...
            ValueError: input contains invalid word

            >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (0,1), (1,0)])
            >>> code.encode([0, 1, 0])  # Code is not uniquely decodable
            Traceback (most recent call last):
            ...
            ValueError: code is not uniquely decodable
        """
        if not self.is_uniquely_decodable():
            raise ValueError("code is not uniquely decodable")
        input = np.asarray(input)
        return parse_fixed_length(input, self.enc_mapping, self.source_block_size)

    def decode(self, input: npt.ArrayLike) -> Array1D[np.integer]:
        r"""
        Decodes a sequence of target symbols using the code, which must be uniquely decodable.

        Warning:
            Decoding for non-prefix-free codes is not implemented yet.

        Parameters:
            input: The sequence of symbols to be decoded. Must be a 1D-array with elements in $\mathcal{Y}$. Also, the sequence must be a concatenation of codewords (i.e., the output of the `encode` method).

        Returns:
            output: The sequence of decoded symbols. It is a 1D-array with elements in $\mathcal{X}$ with a length that is a multiple of $k$.

        Examples:
            >>> code = komm.FixedToVariableCode.from_enc_mapping({
            ...     (0, 0): (0,),
            ...     (0, 1): (1, 1),
            ...     (1, 0): (1, 0, 0),
            ...     (1, 1): (1, 0, 1),
            ... })

            >>> code.decode([1, 1, 0])
            array([0, 1, 0, 0])

            >>> code.decode([0, 0, 1])  # Not a concatenation of codewords
            Traceback (most recent call last):
            ...
            ValueError: input contains invalid word

            >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (0,1), (1,0)])
            >>> code.decode([0, 1, 0])  # Code is not uniquely decodable
            Traceback (most recent call last):
            ...
            ValueError: code is not uniquely decodable
        """
        if not self.is_uniquely_decodable():
            raise ValueError("code is not uniquely decodable")
        if not self.is_prefix_free():
            raise NotImplementedError(
                "decoding for non-prefix-free codes is not implemented yet"
            )
        input = np.asarray(input)
        return parse_prefix_free(input, self._inv_enc_mapping, allow_incomplete=False)



from dataclasses import dataclass
from typing import Literal

import numpy as np
import numpy.typing as npt

from .. import abc
from .._error_control_convolutional.TerminatedConvolutionalCode import (
    TerminatedConvolutionalCode,
)
from .._util.bit_operations import int_to_bits
from .._util.decorators import blockwise, vectorize, with_pbar
from .util import get_pbar


@dataclass
class ViterbiDecoder(abc.BlockDecoder[TerminatedConvolutionalCode]):
    r"""
    Viterbi decoder for [terminated convolutional codes](/ref/TerminatedConvolutionalCode). For more details, see <cite>LC04, Sec. 12.1</cite>.

    Parameters:
        code: The terminated convolutional code to be used for decoding.
        input_type: The type of the input. Either `'hard'` or `'soft'`. Default is `'hard'`.

    Notes:
        - Input type: `hard` (bits) or `soft` (either L-values or channel outputs).
        - Output type: `hard` (bits).
    """

    code: TerminatedConvolutionalCode
    input_type: Literal["hard", "soft"] = "hard"

    def __post_init__(self) -> None:
        if self.code.mode == "tail-biting":
            raise NotImplementedError(
                "Viterbi algorithm not implemented for 'tail-biting'"
            )
        if self.input_type == "hard":
            self._metric_function = self._metric_function_hard
        elif self.input_type == "soft":
            self._metric_function = self._metric_function_soft
        else:
            raise ValueError("input_type must be 'hard' or 'soft'")
        self._fsm = self.code.convolutional_code.finite_state_machine()
        self._post_process_output = self.code.strategy.viterbi_post_process_output
        n = self.code.convolutional_code.num_output_bits
        self._cache_bit = int_to_bits(range(2**n), width=n)
        self._initial_metrics = np.full(self._fsm.num_states, fill_value=np.inf)
        self._initial_metrics[0] = 0.0

    def _metric_function_hard(self, y: int, z: float) -> float:
        return float(np.count_nonzero(self._cache_bit[y] != z))

    def _metric_function_soft(self, y: int, z: int) -> float:
        return np.dot(self._cache_bit[y], z)

    def decode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer | np.floating]:
        r"""
        Examples:
            >>> code = komm.TerminatedConvolutionalCode(
            ...     convolutional_code=komm.ConvolutionalCode([[0b111, 0b101]]),
            ...     num_blocks=4,
            ...     mode="direct-truncation",
            ... )

            >>> decoder = komm.ViterbiDecoder(code, input_type="hard")
            >>> decoder.decode([1, 1, 1, 1, 1, 0, 0, 0])
            array([1, 0, 0, 0])

            >>> decoder = komm.ViterbiDecoder(code, input_type="soft")
            >>> decoder.decode([-0.7, -0.5, -0.8, -0.6, -1.1, +0.4, +0.9, +0.8])
            array([1, 0, 0, 0])
        """
        k = self.code.convolutional_code.num_input_bits
        n = self.code.convolutional_code.num_output_bits

        @blockwise(self.code.length)
        @vectorize
        @with_pbar(get_pbar(np.size(input) // self.code.length, "Viterbi"))
        def decode(r: npt.NDArray[np.integer]):
            xs_hat, final_metrics = self._fsm.viterbi(
                observed=r.reshape(-1, n),
                metric_function=self._metric_function,
                initial_metrics=self._initial_metrics,
            )
            x_hat = self._post_process_output(xs_hat, final_metrics)
            u_hat = int_to_bits(x_hat, width=k).ravel()
            return u_hat

        return decode(input)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import cache, cached_property
from typing import Literal

import numpy as np
import numpy.typing as npt
from numpy.linalg import matrix_power

from .. import abc
from .._util.decorators import blockwise, vectorize
from .._util.matrices import null_matrix, pseudo_inverse
from ..types import Array1D, Array2D

TerminationMode = Literal["direct-truncation", "zero-termination", "tail-biting"]


@dataclass(eq=False)
class TerminatedConvolutionalCode(abc.BlockCode):
    r"""
    Terminated convolutional code. It is a [linear block code](/ref/BlockCode) obtained by terminating a $(n_0, k_0)$ [convolutional code](/ref/ConvolutionalCode). A total of $h$ information blocks (each containing $k_0$ information bits) is encoded. The dimension of the resulting block code is thus $k = h k_0$; its length depends on the termination mode employed. There are three possible termination modes:

    - **Direct truncation**. The encoder always starts at state $0$, and its output ends immediately after the last information block. The encoder may not necessarily end in state $0$. The resulting block code will have length $n = h n_0$.

    - **Zero termination**. The encoder always starts and ends at state $0$. To achieve this, a sequence of $k \mu$ tail bits is appended to the information bits, where $\mu$ is the memory order of the convolutional code. The resulting block code will have length $n = (h + \mu) n_0$.

    - **Tail-biting**. The encoder always starts and ends at the same state. To achieve this, the initial state of the encoder is chosen as a function of the information bits. The resulting block code will have length $n = h n_0$.

    For more details, see <cite>LC04, Sec. 12.7</cite> and <cite>WBR01</cite>.

    Parameters:
        convolutional_code: The convolutional code to be terminated.

        num_blocks: The number $h$ of information blocks.

        mode: The termination mode. It must be one of `'direct-truncation'` | `'zero-termination'` | `'tail-biting'`. The default value is `'zero-termination'`.

    Examples:
        >>> code = komm.TerminatedConvolutionalCode(
        ...     convolutional_code=komm.ConvolutionalCode([[0b1, 0b11]]),
        ...     num_blocks=3,
        ...     mode='direct-truncation',
        ... )
        >>> (code.length, code.dimension, code.redundancy)
        (6, 3, 3)
        >>> code.generator_matrix
        array([[1, 1, 0, 1, 0, 0],
               [0, 0, 1, 1, 0, 1],
               [0, 0, 0, 0, 1, 1]])
        >>> code.minimum_distance()
        2

        >>> code = komm.TerminatedConvolutionalCode(
        ...     convolutional_code=komm.ConvolutionalCode([[0b1, 0b11]]),
        ...     num_blocks=3,
        ...     mode='zero-termination',
        ... )
        >>> (code.length, code.dimension, code.redundancy)
        (8, 3, 5)
        >>> code.generator_matrix
        array([[1, 1, 0, 1, 0, 0, 0, 0],
               [0, 0, 1, 1, 0, 1, 0, 0],
               [0, 0, 0, 0, 1, 1, 0, 1]])
        >>> code.minimum_distance()
        3

        >>> code = komm.TerminatedConvolutionalCode(
        ...     convolutional_code=komm.ConvolutionalCode([[0b1, 0b11]]),
        ...     num_blocks=3,
        ...     mode='tail-biting',
        ... )
        >>> (code.length, code.dimension, code.redundancy)
        (6, 3, 3)
        >>> code.generator_matrix
        array([[1, 1, 0, 1, 0, 0],
               [0, 0, 1, 1, 0, 1],
               [0, 1, 0, 0, 1, 1]])
        >>> code.minimum_distance()
        3
    """

    convolutional_code: abc.ConvolutionalCode
    num_blocks: int
    mode: TerminationMode = "zero-termination"

    def __post_init__(self):
        if not self.mode in TerminationMode.__args__:
            raise ValueError(
                f"mode '{self.mode}' is unknown\n"
                f"supported termination modes: {set(TerminationMode.__args__)}"
            )
        self.strategy = {
            "direct-truncation": DirectTruncation,
            "zero-termination": ZeroTermination,
            "tail-biting": TailBiting,
        }[self.mode](self.convolutional_code, self.num_blocks)

    @cached_property
    def length(self) -> int:
        return self.strategy.codeword_length()

    @cached_property
    def dimension(self) -> int:
        return self.num_blocks * self.convolutional_code.num_input_bits

    @cached_property
    def redundancy(self) -> int:
        return self.length - self.dimension

    @cached_property
    def rate(self) -> float:
        return super().rate

    @cached_property
    def generator_matrix(self) -> Array2D[np.integer]:
        return self.encode(np.eye(self.dimension))

    @cached_property
    def generator_matrix_right_inverse(self) -> Array2D[np.integer]:
        return pseudo_inverse(self.generator_matrix)

    @cached_property
    def check_matrix(self) -> Array2D[np.integer]:
        return null_matrix(self.generator_matrix)

    def encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        @blockwise(self.dimension)
        @vectorize
        def encode(u: npt.NDArray[np.integer]) -> npt.NDArray[np.integer]:
            v, _ = self.convolutional_code.encode_with_state(
                input=self.strategy.pre_process_input(u),
                initial_state=self.strategy.initial_state(u),
            )
            return v

        return encode(input)

    def project_word(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        return super().project_word(input)

    def inverse_encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        return super().inverse_encode(input)

    def check(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        return super().check(input)

    @cache
    def codewords(self) -> Array2D[np.integer]:
        return super().codewords()

    @cache
    def codeword_weight_distribution(self) -> Array1D[np.integer]:
        return super().codeword_weight_distribution()

    @cache
    def minimum_distance(self) -> int:
        return super().minimum_distance()

    @cache
    def coset_leaders(self) -> Array2D[np.integer]:
        return super().coset_leaders()

    @cache
    def coset_leader_weight_distribution(self) -> Array1D[np.integer]:
        return super().coset_leader_weight_distribution()

    @cache
    def packing_radius(self) -> int:
        return super().packing_radius()

    @cache
    def covering_radius(self) -> int:
        return super().covering_radius()


@dataclass
class TerminationStrategy(ABC):
    convolutional_code: abc.ConvolutionalCode
    num_blocks: int

    @abstractmethod
    def pre_process_input(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]: ...

    @abstractmethod
    def initial_state(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]: ...

    @abstractmethod
    def codeword_length(self) -> int: ...

    @abstractmethod
    def viterbi_post_process_output(
        self, xs_hat: npt.NDArray[np.integer], final_metrics: npt.NDArray[np.floating]
    ) -> npt.NDArray[np.integer]: ...

    @abstractmethod
    def bcjr_initial_final_distributions(
        self, num_states: int
    ) -> tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]]: ...

    @abstractmethod
    def bcjr_post_process_output(
        self, posteriors: npt.NDArray[np.floating]
    ) -> npt.NDArray[np.floating]: ...


class DirectTruncation(TerminationStrategy):
    def pre_process_input(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        return np.asarray(input)

    def initial_state(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        σ = self.convolutional_code.degree
        return np.zeros(σ, dtype=int)

    def codeword_length(self) -> int:
        h = self.num_blocks
        n0 = self.convolutional_code.num_output_bits
        return h * n0

    def viterbi_post_process_output(
        self, xs_hat: npt.NDArray[np.integer], final_metrics: npt.NDArray[np.floating]
    ) -> npt.NDArray[np.integer]:
        s_hat = np.argmin(final_metrics)
        x_hat = xs_hat[:, s_hat]
        return x_hat

    def bcjr_initial_final_distributions(
        self, num_states: int
    ) -> tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]]:
        initial_distribution = np.eye(1, num_states, 0)
        final_distribution = np.ones(num_states) / num_states
        return initial_distribution, final_distribution

    def bcjr_post_process_output(
        self, posteriors: npt.NDArray[np.floating]
    ) -> npt.NDArray[np.floating]:
        return posteriors


class ZeroTermination(TerminationStrategy):
    @cached_property
    def _tail_projector(self) -> npt.NDArray[np.integer]:
        # See [WBR01, eq. (3)]. Set x_0 = x_t = 0, and t = h + μ.
        h = self.num_blocks
        μ = self.convolutional_code.memory_order
        σ = self.convolutional_code.degree
        A_mat, B_mat, _, _ = self.convolutional_code.state_space_representation()
        A_pow = [np.eye(σ, dtype=int)]
        for _ in range(1, h + μ):
            A_pow.append((A_pow[-1] @ A_mat) % 2)
        M_info = np.vstack([B_mat @ A_pow[j] % 2 for j in range(h + μ - 1, μ - 1, -1)])
        M_tail = np.vstack([B_mat @ A_pow[j] % 2 for j in range(μ - 1, -1, -1)])
        return M_info @ pseudo_inverse(M_tail) % 2

    def pre_process_input(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        input = np.asarray(input)
        tail = input @ self._tail_projector % 2
        return np.concatenate([input, tail])

    def initial_state(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        σ = self.convolutional_code.degree
        return np.zeros(σ, dtype=int)

    def codeword_length(self) -> int:
        h = self.num_blocks
        n0 = self.convolutional_code.num_output_bits
        μ = self.convolutional_code.memory_order
        return (h + μ) * n0

    def viterbi_post_process_output(
        self, xs_hat: npt.NDArray[np.integer], final_metrics: npt.NDArray[np.floating]
    ) -> npt.NDArray[np.integer]:
        μ = self.convolutional_code.memory_order
        x_hat = xs_hat[:, 0][:-μ]
        return x_hat

    def bcjr_initial_final_distributions(
        self, num_states: int
    ) -> tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]]:
        initial_distribution = np.eye(1, num_states, 0)
        final_distribution = np.eye(1, num_states, 0)
        return initial_distribution, final_distribution

    def bcjr_post_process_output(
        self, posteriors: npt.NDArray[np.floating]
    ) -> npt.NDArray[np.floating]:
        μ = self.convolutional_code.memory_order
        return posteriors[:-μ]


class TailBiting(TerminationStrategy):
    @cached_property
    def _zs_multiplier(self) -> npt.NDArray[np.integer]:
        # See [WBR01, eq. (4)].
        h = self.num_blocks
        σ = self.convolutional_code.degree
        A_mat, _, _, _ = self.convolutional_code.state_space_representation()
        return pseudo_inverse((matrix_power(A_mat, h) + np.eye(σ, dtype=int)) % 2)

    def pre_process_input(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        return np.asarray(input)

    def initial_state(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        σ = self.convolutional_code.degree
        zero_state = np.zeros(σ, dtype=int)
        _, state = self.convolutional_code.encode_with_state(input, zero_state)
        return state @ self._zs_multiplier % 2

    def codeword_length(self) -> int:
        h = self.num_blocks
        n0 = self.convolutional_code.num_output_bits
        return h * n0

    def viterbi_post_process_output(
        self, xs_hat: npt.NDArray[np.integer], final_metrics: npt.NDArray[np.floating]
    ) -> npt.NDArray[np.integer]:
        raise NotImplementedError

    def bcjr_initial_final_distributions(
        self, num_states: int
    ) -> tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]]:
        raise NotImplementedError

    def bcjr_post_process_output(
        self, posteriors: npt.NDArray[np.floating]
    ) -> npt.NDArray[np.floating]:
        raise NotImplementedError



from functools import cache, cached_property

import numpy as np
import numpy.typing as npt

from .. import abc
from .._algebra.BinaryPolynomial import BinaryPolynomial
from .._algebra.BinaryPolynomialFraction import BinaryPolynomialFraction
from .._finite_state_machine.MealyMachine import MealyMachine
from .._util.format import format_list_no_quotes as fmt
from ..types import Array1D, Array2D


class LowRateConvolutionalCode(abc.ConvolutionalCode):
    r"""
    Low-rate convolutional encoder. It is an $(n, 1)$ non-recursive non-systematic [convolutional encoder](/ref/ConvolutionalCode) defined by a single *generator row* $g(D) \in \mathbb{F}_2[D]^n$ and realized in controllable canonical form.

    Parameters:
        g_row: The generator row $g(D)$ of the encoder. Must be an $n$-vector whose entries are either [binary polynomials](/ref/BinaryPolynomial) or integers to be converted to the former.

    Examples:
        Consider the low-rate convolutional encoder with $(n, k, \sigma) = (2, 1, 6)$ depicted below.

        <figure markdown>
        ![Convolutional encoder for low-rate (2, 1, 6) code.](/fig/cc_low_rate_2_1_6.svg)
        </figure>

        Its generator row is given by
        $$
            g(D) =
            \begin{bmatrix}
                D^6 + D^3 + D^2 + D + 1  &&  D^6 + D^5 + D^3 + D^2 + 1
            \end{bmatrix}.
        $$

            >>> code = komm.LowRateConvolutionalCode([0b1001111, 0b1101101])
            >>> code = komm.LowRateConvolutionalCode([0o117, 0o155])

    Please refer to [the table of optimal low-rate convolutional codes](/res/convolutional-codes/#low-rate).
    """

    g_row: list[BinaryPolynomial]

    def __init__(self, g_row: npt.ArrayLike) -> None:
        g_row = np.asarray(g_row, dtype=int)
        if g_row.ndim != 1:
            raise ValueError("'g_row' must be a 1-dimensional array")
        self.g_row = [BinaryPolynomial(x) for x in g_row]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({fmt(self.g_row)})"

    @cached_property
    def num_input_bits(self) -> int:
        r"""
        Examples:
            >>> code = komm.LowRateConvolutionalCode([0o117, 0o155])
            >>> code.num_input_bits
            1
        """
        return 1

    @cached_property
    def num_output_bits(self) -> int:
        r"""
        Examples:
            >>> code = komm.LowRateConvolutionalCode([0o117, 0o155])
            >>> code.num_output_bits
            2
        """
        return len(self.g_row)

    @cached_property
    def degree(self) -> int:
        r"""
        For a low-rate convolutional encoder realized in controllable canonical form, the degree $\sigma$ is the maximum degree of the polynomials in the generator row $g(D)$.

        Examples:
            >>> code = komm.LowRateConvolutionalCode([0o117, 0o155])
            >>> code.degree
            6
        """
        return max(p.degree for p in self.g_row)

    @cached_property
    def generator_matrix(self) -> Array2D[np.object_]:
        r"""
        For a low-rate convolutional code, it is given by
        $$
            G(D) = \big[ ~ g(D) ~ \big].
        $$

        Examples:
            >>> code = komm.LowRateConvolutionalCode([0o117, 0o155])
            >>> for row in code.generator_matrix:
            ...     print("[" + ", ".join(str(x) for x in row) + "]")
            [0b1001111/0b1, 0b1101101/0b1]
        """
        n, k = self.num_output_bits, self.num_input_bits
        G_mat = np.empty((k, n), dtype=object)
        for j in range(n):
            p = BinaryPolynomialFraction(self.g_row[j])
            G_mat[0, j] = p
        return G_mat

    @cached_property
    def constraint_lengths(self) -> Array1D[np.integer]:
        r"""
        Examples:
            >>> code = komm.LowRateConvolutionalCode([0o117, 0o155])
            >>> code.constraint_lengths
            array([6])
        """
        return np.array([self.degree], dtype=int)

    @cached_property
    def overall_constraint_length(self) -> int:
        r"""
        Examples:
            >>> code = komm.LowRateConvolutionalCode([0o117, 0o155])
            >>> code.overall_constraint_length
            6
        """
        return super().overall_constraint_length

    @cached_property
    def memory_order(self) -> int:
        r"""
        Examples:
            >>> code = komm.LowRateConvolutionalCode([0o117, 0o155])
            >>> code.memory_order
            6
        """
        return super().memory_order

    @cache
    def state_space_representation(
        self,
    ) -> tuple[
        Array2D[np.integer],
        Array2D[np.integer],
        Array2D[np.integer],
        Array2D[np.integer],
    ]:
        r"""
        Examples:
            >>> code = komm.LowRateConvolutionalCode([0o117, 0o155])
            >>> A_mat, B_mat, C_mat, D_mat = code.state_space_representation()
            >>> A_mat
            array([[0, 1, 0, 0, 0, 0],
                   [0, 0, 1, 0, 0, 0],
                   [0, 0, 0, 1, 0, 0],
                   [0, 0, 0, 0, 1, 0],
                   [0, 0, 0, 0, 0, 1],
                   [0, 0, 0, 0, 0, 0]])
            >>> B_mat
            array([[1, 0, 0, 0, 0, 0]])
            >>> C_mat
            array([[1, 0],
                   [1, 1],
                   [1, 1],
                   [0, 0],
                   [0, 1],
                   [1, 1]])
            >>> D_mat
            array([[1, 1]])
        """
        σ = self.degree
        beta = np.array([p.coefficients(σ + 1) for p in self.g_row]).T
        A_mat = np.eye(σ, σ, k=1, dtype=int)
        B_mat = np.eye(1, σ, dtype=int)
        C_mat = beta[1:]
        D_mat = np.array([beta[0]])
        return A_mat, B_mat, C_mat, D_mat

    @cache
    def finite_state_machine(self) -> MealyMachine:
        r"""
        Examples:
            >>> code = komm.LowRateConvolutionalCode([0o117, 0o155])
            >>> fsm = code.finite_state_machine()
            >>> (fsm.num_input_symbols, fsm.num_output_symbols, fsm.num_states)
            (2, 4, 64)
        """
        return super().finite_state_machine()

    @cache
    def is_catastrophic(self) -> bool:
        """
        Examples:
            >>> code = komm.LowRateConvolutionalCode([0o117, 0o155])
            >>> code.is_catastrophic()
            False
        """
        return super().is_catastrophic()

    @cache
    def free_distance(self) -> int:
        r"""
        Examples:
            >>> code = komm.LowRateConvolutionalCode([0o117, 0o155])
            >>> code.free_distance()
            10
        """
        return super().free_distance()

    def encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.LowRateConvolutionalCode([0o117, 0o155])
            >>> code.encode([1, 1, 1, 1])
            array([1, 1, 0, 1, 1, 0, 0, 1])
        """
        return super().encode(input)

    def encode_with_state(
        self,
        input: npt.ArrayLike,
        initial_state: npt.ArrayLike,
    ) -> tuple[npt.NDArray[np.integer], npt.NDArray[np.integer]]:
        r"""
        Examples:
            >>> code = komm.LowRateConvolutionalCode([0o117, 0o155])
            >>> code.encode_with_state([1, 1, 1, 1], [0, 0, 0, 0, 0, 0])
            (array([1, 1, 0, 1, 1, 0, 0, 1]), array([1, 1, 1, 1, 0, 0]))
            >>> code.encode_with_state([1, 1, 1, 1], [1, 1, 1, 1, 0, 0])
            (array([0, 1, 0, 0, 1, 1, 1, 1]), array([1, 1, 1, 1, 1, 1]))
        """
        return super().encode_with_state(input, initial_state)
